##############################################################################
# german_page.py
#
# Christine Pranzas, 2007
#
##############################################################################

## german_page
## is a page for german lute tabulature
from gamera import toolkit
from math import sqrt
import random
import sys
from german_page import *
import plugins
from gamera.core import *
from gamera.args import *
from gamera.gui import gui
from gamera import toolkit
from gamera.toolkits.otr.plugins import *
from gamera.core import *
from gamera.gui import has_gui
from gamera.plugins.image_utilities import union_images
if has_gui.has_gui:
    from gamera.gui import var_name
    import wx


class GermanLutePage:
##############################################################################
    """Performs layout analysis on a german page. The layout analysis assumes
that lines of tablature are separated by horizontal lines.

The complete layout analysis is done in the constructor. The results are
stored in the following properties:

  *line_height*, *character_height*, *distance_between_two_lines*
    characteristic page dimensions

  *line_groups*
    Array of detected lines. Each line is itself a list of all vertical runs
    belonging to the line.

  *straight_lines*
    List of the parameters (m,b) for all least square fitted lines.

For interactive testing, a *GermanLutePage* object can also be created from 
the context menu of the OTR toolkit icon in the Gamera GUI. This will create
an icon representing the *GermanLutePage* object, and from its context menu
you can display images showing the found separator lines or bars.
"""

    class GermanLutePageIcon(toolkit.CustomIcon):
        def get_icon():
            import germanlutepage_icon as pp_icon
            return toolkit.CustomIcon.to_icon(pp_icon.getBitmap())
        get_icon = staticmethod(get_icon)
    
        def check(data):
            return isinstance(data, GermanLutePage)

        check = staticmethod(check)
        
        def right_click(self, parent, event, shell):
            
            self._shell=shell
            x, y=event.GetPoint()
            #global menu
            self.menu=wx.Menu()
            self.parent=parent
            # create the 'Info' menu point and insert it to
            # the "right click menu"
            info_menu=wx.Menu()
            self.menu.AppendMenu(wx.NewId(), "Info", info_menu)

            info_menu.Append(wx.NewId(), "Class instance: %s"\
                    % str(self.data.__class__).\
                    split(".")[-1])
            info_menu.AppendSeparator()
			
            # call characteristic_dimensions function
            # and add the result in 'Info' menu
            chara_values=[]
            chara_values=[self.data.line_height,self.data.character_height,self.data.distance_between_two_lines]
            if len(chara_values) > 0:
                info_menu.Append(wx.NewId(),"Line heigth: %d" % chara_values[0])
                info_menu.Append(wx.NewId(),"Character_height: %d" % chara_values[1])
                info_menu.Append(wx.NewId(),"Distance_between_two_lines: %d" % chara_values[2])
 

            # create the rest of the menu
            self.menu.AppendSeparator()
            index=wx.NewId()
            self.menu.Append(index, "Show original image")
            EVT_MENU(parent, index, self.call_show_original_image)
            index=wx.NewId()
            #if self.removed_lyrics:
            self.menu.Append(index, "Show image")
            EVT_MENU(parent, index, self.call_show_image)
            index=wx.NewId()
            self.menu.Append(index, "Copy from original image")
            EVT_MENU(parent, index, self.call_copy_original_image)
            index=wx.NewId()
            self.menu.Append(index, "Copy from image")
            EVT_MENU(parent, index, self.call_copy_image)
  ##           index=wx.NewId()
##             self.menu.Append(index, "Find lines")
##             EVT_MENU(parent, index, self.call_find_lines)
            index=wx.NewId()
            self.menu.Append(index, "Show lines")
            EVT_MENU(parent, index, self.call_show_lines)
            index=wx.NewId()
            self.menu.Append(index, "Show bars")
            EVT_MENU(parent, index, self.call_show_bars)
            
 
          
            parent.PopupMenu(self.menu, wx.Point(x, y))
    
    	##################################################################
    	# further implementiation of called functions
    
        def double_click(self):
        	return "%s.image.display()" % self.label

##         def call_find_lines(self, event):
##         	self._shell.run("%s.find_lines()" % self.label)

        def call_show_lines(self, event):
            self._shell.run("%s.gui_show_lines()" % self.label)

        def call_show_bars(self, event):
            self._shell.run("%s.gui_show_bars()" % self.label)
       
        def call_show_original_image(self, event):
            self._shell.run("%s.original_image.display()" % self.label)
            
        def call_show_image(self, event):
            self._shell.run("%s.image.display()" % self.label)
		
        def call_copy_original_image(self, event):
            img_name=var_name.get(self.label + "_image",\
                                  self._shell.locals)

            if img_name is not None:
                self._shell.run("%s = %s.original_image.image_copy()"\
                        % (img_name, self.label))
                
        def call_copy_image(self, event):
            img_name=var_name.get(self.label + "_image",\
                                  self._shell.locals)

            if img_name is not None:
                self._shell.run("%s = %s.image.image_copy()"\
                        % (img_name, self.label))


    
    ######################################################################
    # constructor
    #
    def __init__(self, img,bar_detection=True,clean=False,log_file=None,debug=0):
        """The constructor already does all the layout analysis. Consequently
there is usually no need to call any other methods of *GermanLutePage*.

Signature:

  ``__init__(image, bar_detection=True, clean=False, log_file=None, debug=0)``

with

  *image*:
    Onebit or greyscale image to be processed.

  *bar_detection*:
    Whether bars are to be detected heuristically.

  *clean*:
    Whether gaps in fragmented separator lines shall be searched for
    additional line fragments.

  *log_file*:
    File name of a log file for debugging. When *None*, no logfile is written.

  *debug*:
    Debug level (i.e. verbosity level). Can be 0,1 or 2.

"""

        if img.data.pixel_type!=ONEBIT:
            self.original_image=img.to_onebit()
        else:
            self.original_image=img.image_copy()

        if img.ul_x!=0 or img.ul_y!=0:
            img2=Image((0,0),img.size,ONEBIT)
            img2.or_image(self.image,True)
            self.image=img2

        self.image=self.original_image.to_rgb()
        self.clean=clean
        self.log_file=log_file
        self.line_height=0
        self.debug=debug
        self.character_height=0
        self.lines=self.find_lines()
        self.straight_lines=self.get_linear_equation()
        self.bars=[]
        self.bar_detection=bar_detection
        
        if bar_detection:
            self.bars=self.find_bars()
            #self.cc_bars=self.get_bar_ccs()
        if has_gui.has_gui:
            self.GermanLutePageIcon.register()


            
            

##--------------------------------------------------------------------------------

    def get_parameters(self,line):
        # get_parameters uses the linear regression to get the gradient and the y axis intercept of each line

        # with least squares fit
        su_xi=0.0
        su_yi=0.0
        su_xi_yi=0.0
        su_xi_2=0.0
        N=0
        for rect in line:
        #    print "r",rect.center_x,rect.center_y
            su_xi+=rect.center_x
            su_yi+=rect.center_y
            su_xi_yi+=rect.center_x*rect.center_y
            su_xi_2+=rect.center_x*rect.center_x
            N+=1
        if N==1:
            m=0
        else:
            m=(su_xi_yi-su_xi*su_yi/N)/(su_xi_2-su_xi*su_xi/N)
        b=(su_yi-m*su_xi)/N
        return m,b

##--------------------------------------------------------------------------------

    def check_overlap(self,belongs_to_line,long_line):
        #check if two lines overlapp. If they overlap delete the line which is farer away
        
        new_line=long_line
        for n,belong in enumerate( belongs_to_line):
            new_line=new_line+belong
        m,b=self.get_parameters(new_line)

     ##    for x in range(image.width):
    ##         y=m*x+b
    ##         if y<0 or y >= rgb_image.height:
    ##             continue
    ##         rgb_image.set(Point(x,y),RGBPixel(255,0,0))


        short_idx=0
        while short_idx<len(belongs_to_line):
            short_sec_idx=0
            while short_sec_idx<len(belongs_to_line):
                if short_idx==short_sec_idx:
                    short_sec_idx+=1
                    continue

                left=max(belongs_to_line[short_idx][0].center_x,belongs_to_line[short_sec_idx][0].center_x)
                right=min(belongs_to_line[short_idx][-1].center_x,belongs_to_line[short_sec_idx][-1].center_x)
                overlapped=right-left



                if overlapped>self.line_height:
                    sum1=0.0
                    for first in belongs_to_line[short_idx]:
                        sum1+=first.center_y
                    sum1=sum1/len(belongs_to_line[short_idx])
                    sum2=0.0
                    for second in belongs_to_line[short_sec_idx]:
                        sum2+=second.center_y

                    x_1=(belongs_to_line[short_idx][0].center_x+belongs_to_line[short_idx][-1].center_x)/2
                    x_2=(belongs_to_line[short_sec_idx][0].center_x+belongs_to_line[short_sec_idx][-1].center_x)/2            
                    sum2=sum2/len(belongs_to_line[short_sec_idx])
                    dist1=abs(sum1-(m*x_1+b))
                    dist2=abs(sum2-(m*x_2+b))

                    if (dist1<dist2):
                        belongs_to_line.remove(belongs_to_line[short_sec_idx])
                        short_idx=-1
                        break
                    elif dist1>dist2:
                        belongs_to_line.remove(belongs_to_line[short_idx])
                        short_idx=-1
                        break
                    else:
                        if len(belongs_to_line[short_sec_idx])<len(belongs_to_line[short_idx]):
                            belongs_to_line.remove(belongs_to_line[short_sec_idx])
                        elif len(belongs_to_line[short_sec_idx])>len(belongs_to_line[short_idx]) :
                            belongs_to_line.remove(belongs_to_line[short_idx])
                        else:
                            belongs_to_line.remove(belongs_to_line[short_idx])

                        short_idx=-1
                        break        
                else:
                    short_sec_idx+=1

            short_idx+=1
        return belongs_to_line
    
##--------------------------------------------------------------------------------
    def find_lines(self):
        #find the horizontal lines in an image.
        grey_image=self.image.to_onebit()
        rgb_image = grey_image.to_rgb()
        image = grey_image.to_onebit()
        onebit_image = grey_image.to_onebit()
        image_width=rgb_image.width


        ##     #try to get character_height and flag height

        ccs=onebit_image.cc_analysis()
        for cc in ccs:
            if cc.height<7 or cc.width <7:
                cc.fill_white()

        characters=onebit_image.cc_analysis()
        characters.sort(lambda a,b:cmp(a.width,b.width))

        #character_width=characters[len(characters)/2].width

        characters.sort(lambda a,b:cmp(a.height,b.height))

        character_height=characters[len(characters)/2].height

        #print "character_heihgt, character_widht",character_height,character_width

        print "character_heihgt",character_height

        self.character_height=character_height

        # filter all runs bigger than charcter_height 
        onebit_image.filter_tall_runs(self.character_height,u'black')


        ccs=onebit_image.cc_analysis()


        #clean image up
        for cc in ccs:
            if float(cc.ncols/cc.nrows)<3.0 : #or cc.ncols<onebit_image.ncols/20:
                cc.fill_white()

        self.line_height = onebit_image.most_frequent_run('black','vertical')

        print "Most frequent run", self.line_height


        #find the root mean square deviation

        #s_x=sqrt( (N*sum(x_i^2,1,N)-sum(x_i,1,N)^2)/(N*(N-1)) )
        sum_1=0.0
        sum_2=0.0
        N=0
        for x_iter in onebit_image.iterate_runs('black','vertical'):
            for run in x_iter:
                sum_1+=run.height*run.height
                sum_2+=run.height
                N+=1

        s_x=sqrt((N*sum_1-sum_2*sum_2)/(N*(N-1)))

        if self.debug>1:
            print "root mean square deviation", s_x                    

        #Get all runs and save those with a height <=self.line_height
        #These are potential candidates for members of one line 
        x_list=[]
        all_bigger_runs=[]
        for x_iter in image.iterate_runs('black','vertical'):
            x_list.append([])
            all_bigger_runs.append([])
            for run in x_iter:
                if run.height<=self.line_height+s_x*2:
                    #print run.width
                    x_list[-1].append(run)
                    #rgb_image.draw_line(run.ul,run.ll,RGBPixel(0,0,255))

                else:
                    all_bigger_runs[-1].append(run)


        #print "searching for lines"


        #Searching for lines
        #-------------------------------------#
        #Go trough the list of potential candidates, and try to find related pixel lines 
        lines=[]
        while True:
            x=0
            #jump over empty array elements
            while x<len(x_list) and len(x_list[x])==0:
                x+=1
            if x>=len(x_list):
                break
            #save all elements belonging to a line in run_list
            #first element of a line
            run_list=[x_list[x][0]]
            #delete already assigned elements
            del x_list[x][0]
            x+=1
            #For each line breaks in range of 2*self.line_height allowed until the next part of the line is found. When after this distance no member of a line is found. The line is complete and the next line is searched.
            while x<len(x_list) and x<run_list[-1].center_x+self.line_height*2:
                found=False
                for run in x_list[x]:
                    #the center_y value between the last element of the line and the new value must be smaller self.line_height*0.75
                    if abs(run.center_y-run_list[-1].center_y)<=self.line_height*0.75:
                        # when there are alternatives, take that one which height is bigger
                        if found==False:
                            found=run
                        else:
                            if run.height>found.height:
                                found=run
                if found!=False:
                    run_list.append(found)
                    x_list[x].remove(found)
                x+=1
            lines.append(run_list)

        #print "classify into long and short lines"


        #classify those lines into long and short lines
        long_lines=[]
        short_lines=[]
        for line in lines:
            #if len(line)>character_height*3:
            if len(line)>image_width/10:
                long_lines.append(line)
            else:
                short_lines.append(line)


        #sort the long_lines in line_groups. Every line_group contains all line elements beloning to one line 
        line_groups=[]
        for line in long_lines:
            group_found=False
            for group in line_groups:
                if abs(line[0].center_y-group[0][0].center_y)<6*self.line_height:
                    group_found=True
                    group.append(line)
            if not group_found:
                line_groups.append([line])
        line_groups.sort(lambda x,y: cmp(x[0][0].center_y,y[0][0].center_y))
        for group in line_groups:
            group.sort(lambda x,y: cmp(x[0].center_x,y[0].center_x))

        #compute the average distance between two lines
    ##     distance_between_two_lines=0
    ##     N=0
    ##     for line_number,line in enumerate(line_groups):
    ##         if line_number==len(line_groups)-1:
    ##             break
    ##         if line_groups[line_number+1][0][0].center_y-line[0][0].center_y>3*character_height:
    ##             #print "dist",line_groups[line_number+1][0][0].center_y-line[0][0].center_y
    ##             distance_between_two_lines+=(line_groups[line_number+1][0][0].center_y-line[0][0].center_y)
    ##             N+=1
    ##     if distance_between_two_lines!=0:
    ##         distance_between_two_lines=distance_between_two_lines/N
    ##     print "distance_between_two_lines",distance_between_two_lines

        distance_between_two_lines_array=[]
        for line_number,line in enumerate(line_groups):
            if line_number==len(line_groups)-1:
                break
            if line_groups[line_number+1][0][0].center_y-line[0][0].center_y>3*self.character_height:
                distance_between_two_lines_array.append(line_groups[line_number+1][0][0].center_y-line[0][0].center_y)
        if distance_between_two_lines_array!=[]:
            distance_between_two_lines_array.sort()
            distance_between_two_lines=distance_between_two_lines_array[(len(distance_between_two_lines_array)-1)/2]
            print "distance_between_two_lines",distance_between_two_lines
        elif character_height>5:
            distance_between_two_lines=character_height*5
        else:
            distance_between_two_lines=0

        self.distance_between_two_lines=distance_between_two_lines

        #print distance_between_two_lines_array
        #check if a linepeace that does not belong to line is wrongly classified
        #----------------------------------------------------------------------
        gr_number=0
        while gr_number<len(line_groups)-1:
            first_line=[]
            for line in line_groups[gr_number]:
                first_line+=line
            m1,b1=self.get_parameters(first_line)

            sec_line=[]
            for line in line_groups[gr_number+1]:
                sec_line+=line
            m2,b2=self.get_parameters(sec_line)

           ## if line_groups[gr_number][0][0].center_x<line_groups[gr_number+1][0][0].center_x:
            if line_groups[gr_number][0][0].center_x<line_groups[gr_number+1][0][0].center_x:
                x=(line_groups[gr_number][-1][-1].center_x+line_groups[gr_number+1][0][0].center_x)/2
            else:
                x=(line_groups[gr_number+1][-1][-1].center_x+line_groups[gr_number][0][0].center_x)/2
            dist=abs((m1 * x+b1)-(m2 * x+b2))

    ##         if len(first_line)>=len(sec_line):
    ##             x=sec_line[0].center_x
    ##         else:
    ##             x=first_line[0].center_x

    ##         dist=abs((m1 * x+b1)-(m2 * x+b2))

            #print "#### steigung",m1,m2,b1,b2
           # print "dist", dist, line_groups[gr_number][0][0].center_x,line_groups[gr_number][0][0].center_y, line_groups[gr_number+1][0][0].center_x,line_groups[gr_number+1][0][0].center_y
            if dist>self.distance_between_two_lines/5 and dist<self.distance_between_two_lines/2:


                len_gr_1=0
                y_mid_1=0
                for group in line_groups[gr_number]:
                    len_gr_1+=len(group)
                    for line in group:
                        y_mid_1+=line.center_y

                len_gr_2=0
                y_mid_2=0
                for group in line_groups[gr_number+1]:
                    len_gr_2+=len(group)
                    for line in group:
                        y_mid_2+=line.center_y

                if len_gr_1>len_gr_2:
                    short_lines+=line_groups[gr_number+1]
                    line_groups.remove(line_groups[gr_number+1])
                    group_number=-1

                elif len_gr_1<=len_gr_2:
                    short_lines+=line_groups[gr_number]
                    line_groups.remove(line_groups[gr_number])
                    group_number=-1
            elif dist<self.distance_between_two_lines/5:
                line_groups[gr_number]=line_groups[gr_number]+line_groups[gr_number+1]
                line_groups.remove(line_groups[gr_number+1])
                #print "################################################++++++++++++++++++++++++++++++++++++++++++++ Append one line to another"
                line_groups[gr_number].sort(lambda x,y: cmp(x[0].center_x,y[0].center_x))
                gr_number=-1


            gr_number+=1

        line_groups.sort(lambda x,y: cmp(x[0][0].center_y,y[0][0].center_y))



    ##     #Ascertian if two long_lines belonging to the same line, are classified as two different line
        gr_number=0
        first=False
        while(gr_number<len(line_groups)):
            groups_to_append=[]
            if gr_number==len(line_groups)-1:


                break

            next_group=line_groups[gr_number+1]

            max_value_of_first_line=0
            min_value_of_second_line=image.height
            for lines in line_groups[gr_number]:
                for line in lines:
                    if line.center_y>max_value_of_first_line:
                        max_value_of_first_line=line.center_y

            for lines in next_group:
                for line in lines:
                    if line.center_y<min_value_of_second_line:
                        min_value_of_second_line=line.center_y

            dist=abs(min_value_of_second_line-max_value_of_first_line)

            if (dist<self.distance_between_two_lines/2):
                line_groups[gr_number]=line_groups[gr_number]+next_group
                line_groups.remove(next_group)
               # print "################################################++++++++++++++++++++++++++++++++++++++++++++ Append one line to another"
                line_groups[gr_number].sort(lambda x,y: cmp(x[0].center_x,y[0].center_x))
                gr_number=-1

            gr_number+=1


    ##     for group in line_groups:
    ##         color=RGBPixel(random.randint(34,224),random.randint(34,224),random.randint(34,224))
    ##         for line in group:
    ## 	    for rect in line:
    ##                 rgb_image.draw_line(rect.ul,rect.ll,color)






        #Verify if two long_lines overlapp in the same line_group. When two long_lines overlapp, delete the shorter one
        gr_number=0
        while gr_number<len(line_groups):
            line_number=0
            while line_number<len(line_groups[gr_number]):
                line_sec_number=0
                while line_sec_number<len(line_groups[gr_number]):
                    #print "hallo",line_sec_number
                    if line_number==line_sec_number:
                        line_sec_number+=1
                        continue
                    left=max(line_groups[gr_number][line_number][0].center_x,line_groups[gr_number][line_sec_number][0].center_x)
                    right=min(line_groups[gr_number][line_number][-1].center_x,line_groups[gr_number][line_sec_number][-1].center_x)
                    overlapped=right-left

                    if overlapped>self.character_height*2:

                        if len(line_groups[gr_number][line_number])>len(line_groups[gr_number][line_sec_number]):
                           # print "Remove sec line"
                           # short_lines+=line_groups[gr_number][line_sec_number]
                            line_groups[gr_number].remove(line_groups[gr_number][line_sec_number])

                            line_number=-1
                            break
                        elif len(line_groups[gr_number][line_number])<=len(line_groups[gr_number][line_sec_number]):
                           # print "Remove line "
                            #short_lines+=line_groups[gr_number][line_number]
                            line_groups[gr_number].remove(line_groups[gr_number][line_number])
                            line_number=-1
                            break


                    else:
                        line_sec_number+=1
                line_number+=1

            line_groups[gr_number].sort(lambda x,y: cmp(x[0].center_x,y[0].center_x))

            gr_number+=1



        ##     #check if a line is missing
    ##++++++++++++++++++++++++++++++++++++
        gr_number=0
        while gr_number<len(line_groups)-1:
            dist=0
    ##         if gr_number==-1:
    ##             dist=abs(line_groups[0][0][0].center_y)
    ##             y_value=(line_groups[0][0][0].center_y)/2

    ##         elif gr_number==(len(line_groups)-1):
    ##             dist=abs(line_groups[-1][0][0].center_y-image.height)
    ##             y_value=(line_groups[gr_number][0][0].center_y+image.height)/2
            #else:
            #print ".......... ",gr_number,len(line_groups)
            dist=abs(line_groups[gr_number][0][0].center_y-line_groups[gr_number+1][0][0].center_y)
            y_value=(line_groups[gr_number][0][0].center_y+line_groups[gr_number+1][0][0].center_y)/2


            if dist>2.5*self.distance_between_two_lines and gr_number!=-1 and gr_number!=len(line_groups)-1:
                print "ERRROR: Two lines not detected!"

            elif dist>1.5*self.distance_between_two_lines:
    ##             lines_to_append=[]
    ##             short_idx=0

    ##             while short_idx<len(short_lines):
    ##                 if len(short_lines[short_idx])>image.width/15 and short_lines[short_idx][0].center_y>y_value-distance_between_two_lines/4 and short_lines[short_idx][0].center_y<y_value+distance_between_two_lines/4:
    ##                     lines_to_append.append(short_lines[short_idx])

    ##                 short_idx+=1
    ##             lines_to_append.sort(lambda x,y: cmp(x[0].center_x,y[0].center_x))
    ##             group_number=-1


                long_lines=[]
                for lines in line_groups[gr_number]:
                    long_lines+=lines
                m,b=self.get_parameters(long_lines)
                #print "hihi",m,b
                #print long_lines
                y_count=y_value-2*self.line_height
                most_pixel=-1
                while y_count<y_value+2*self.line_height:
                    count_lines=[]
    ##                 for x in range(image.width):
    ##                    
    ##                     if rgb_image.get((x,y_count))!=RGBPixel(255,255,255):
    ##                         count+=1
                    for lines in short_lines:
                        if len(lines)<self.line_height*15:
                            continue
                        for rect in lines:
                            y=m*rect.center_x+y_count
                            if rect.ul_y<=y and rect.ll_y>=y:
                                count_lines.append(lines)
                                break
                    count=0
                    for cl in count_lines:
                        for rect in cl:
                            count+=1

                    if count>0 and most_pixel<count:
                        most_pixel=count
                        most_rects=[]
                        for cl in count_lines:
                            most_rects.append(cl)
                    y_count+=1

                if most_pixel!=-1:
      ##               for mr in most_rects:
    ##                     color=RGBPixel(random.randint(34,224),random.randint(34,224),random.randint(34,224))
    ##                     for rect in mr:
    ##                             rgb_image.draw_line(rect.ul,rect.ll,color)

                    line_groups.append(most_rects)
                    line_groups.sort(lambda x,y: cmp(x[0][0].center_y,y[0][0].center_y))           

    ##                 for x in range(image.width):
    ##                     y=m*x+most_y
    ##                     if y<0 or y >= rgb_image.height:
    ##                         continue
    ##                     rgb_image.set(Point(x,y),RGBPixel(255,0,0))

                #rgb_image.draw_line((0,y_value),(image.width,y_value),RGBPixel(255,0,0),5)

               # print "..__-...--.-.-.-.-......--..-.--.--.- distance between two lines to big"



            gr_number+=1


        #berechne Steiung und Achsenabschnitt fuer alle long_lines:
        #least squares fit

        parameter_long_line=[]
        whole_long_line=[]
        for gr_number,group in enumerate(line_groups):
            parameter_long_line.append([])
            whole_long_line.append([])
            for line_number,line in enumerate(group): 
                m,b=self.get_parameters(line)
                parameter_long_line[gr_number].append([m,b])
                whole_long_line[gr_number]+=line
                #draw the linear reggresion line
               ##  for x in range(image.width):
    ##                 y=m*x+b
    ##                 if y<0 or y >= rgb_image.height:
    ##                     continue
    ##                 rgb_image.set(Point(x,y),RGBPixel(255,0,0))


        #Sort the short_lines, into short_groups hinge on the distance of the line_groups. The short_group array contains all short lines who might be potenial candiates for the assigned line_group.
        short_groups=[]
        for group in line_groups:
            short_groups.append([])
        for line in short_lines:
            dist=100000
            gr_idx=-1
            for group in line_groups:
                if abs(line[0].center_y-group[0][0].center_y)<dist:
                    dist=abs(line[0].center_y-group[0][0].center_y)
                    gr_idx=line_groups.index(group)
            short_groups[gr_idx].append(line)

        for group in short_groups:
            group.sort(lambda x,y: cmp(x[0].center_x,y[0].center_x))



        #sort short lines into groups with least squares fit
        overlap=self.line_height*3
        gr_idx=0
        while gr_idx <len(line_groups):
            short_left=[]
            short_right=[]
            short_inbetween=[]
            for line in line_groups[gr_idx]:
                short_inbetween.append([])
            for short_line in short_groups[gr_idx]:
                #check if line is on the left , inbetween or on the right side
                if (short_line[-1].center_x-overlap)<line_groups[gr_idx][0][0].center_x:
                    #check line is on the left side
                    short_left.append(short_line) 
                elif (short_line[0].center_x+overlap)>line_groups[gr_idx][-1][-1].center_x:
                    #line is on the right
                    short_right.append(short_line)
                #check if line is inbetween
                else:
                    for line_number,long_line in enumerate(line_groups[gr_idx]):
                        if line_number==len(line_groups[gr_idx])-1:
                            break
                        if ((long_line[-1].center_x-overlap)<short_line[0].center_x) and  ((short_line[-1].center_x-overlap)<line_groups[gr_idx][line_number+1][0].center_x):
                                short_inbetween[line_number].append(short_line)


            #check inbetween
            #---------------
            belongs_to_right_line=[]
            belongs_to_left_line=[]
            for i_inbetween in range( len(short_inbetween)-1):

                belongs_to_right_line.append([])
                belongs_to_left_line.append([])
                if len(short_inbetween[i_inbetween])==0:
                    continue


                ml=parameter_long_line[gr_idx][i_inbetween][0]
                bl=parameter_long_line[gr_idx][i_inbetween][1]
                mr=parameter_long_line[gr_idx][i_inbetween+1][0]
                br=parameter_long_line[gr_idx][i_inbetween+1][1]
                #print "+++++old m,b ",m,b
                changed=True

                while(changed==True):

                    idx=0
                    short_close_right_distance=[]
                    dist_right=image.height
                    index_right=-1
                    center_x_right=image.width

                    short_close_left_distance=[]
                    dist_left=image.height
                    index_left=-1
                    center_x_left=image.width
                    while idx< len(short_inbetween[i_inbetween]):
                        #print "hallo", idx

                        mid=len(short_inbetween[i_inbetween][idx])/2
                        y_short=short_inbetween[i_inbetween][idx][mid].center_y
                        x_short=short_inbetween[i_inbetween][idx][mid].center_x
                        #y_long = m*x_short+b
                        y_long_left=ml*x_short+bl
                        y_long_right=mr*x_short+br

                        if abs(y_short-y_long_left)<self.line_height:
                            #print "Short lines attached:",len(short_)
                            if index_left==-1 or dist_left>abs(y_short-y_long_left) or dist_left==abs(y_short-y_long_left) and short_inbetween[i_inbetween][idx][0].center_x<center_x_left:
                                index_left=idx
                                dist_left=int(abs(y_short-y_long_left))
                                center_x_left=short_inbetween[i_inbetween][idx][0]
                        elif abs(y_short-y_long_right)<self.line_height:
                            if index_right==-1 or dist_right>abs(y_short-y_long_right) or dist_right==abs(y_short-y_long_right) and short_inbetween[i_inbetween][idx][-1].center_x>center_x_right:
                                index_right=idx

                                dist_right=int(abs(y_short-y_long_right))
                                center_x_right=short_inbetween[i_inbetween][idx][-1]
                        idx+=1
                    if index_left==index_right:
                        index_right=-1
                    if index_left!=-1:  
                        belongs_to_left_line[i_inbetween].append(short_inbetween[i_inbetween][index_left])
                        short_inbetween[i_inbetween].remove(short_inbetween[i_inbetween][index_left])
                        if index_left<index_right:
                            index_right-=1
                        new_line_left=line_groups[gr_idx][i_inbetween]
                        for n,belong in enumerate( belongs_to_left_line[i_inbetween]):
                            new_line_left=new_line_left+belong
                        ml,bl=self.get_parameters(new_line_left)
                    if index_right!=-1:
                        belongs_to_right_line[i_inbetween].append(short_inbetween[i_inbetween][index_right])
                        short_inbetween[i_inbetween].remove(short_inbetween[i_inbetween][index_right])
                        new_line_right=line_groups[gr_idx][i_inbetween]
                        for n,belong in enumerate( belongs_to_right_line[i_inbetween]):
                            new_line_right=new_line_right+belong
                        mr,br=self.get_parameters(new_line_right)


                    if index_left==-1 and index_right==-1: 
                        changed=False


            #CHECK if lines overlapp, therefore append the right_lines to the left_lines
            for i_inb in range(len(short_inbetween)-1):
                belongs_to_left_line[i_inb]=self.check_overlap(belongs_to_right_line[i_inb]+belongs_to_left_line[i_inb],whole_long_line[gr_idx])


            for i_inb in range(len(short_inbetween)-1):
                #if len(belongs_to_left_line[i_inb])>0:
                #line_groups[gr_idx]=line_groups[gr_idx]+belongs_to_left_line[i_inb]+belongs_to_right_line[i_inb]
                line_groups[gr_idx]=line_groups[gr_idx]+belongs_to_left_line[i_inb]



            line_groups[gr_idx].sort(lambda x,y: cmp(x[0].center_x,y[0].center_x))



            #check left side
            #---------------
            if short_left!=[]:
                #print "len(short_left)",len(short_left)

                belongs_to_line=[]
                #m,b=self.get_parameters(whole_long_line[gr_idx])
                m=parameter_long_line[gr_idx][0][0]
                b=parameter_long_line[gr_idx][0][1]
                #print "+++++old m,b ",m,b
                changed=True
                while(changed==True):
                    idx=0
                    short_close_distance=[]
                    dist=image.height
                    index=-1
                    center_x=image.width
                    while idx< len(short_left):
                        mid=len(short_left[idx])/2
                        y_short=short_left[idx][mid].center_y
                        x_short=short_left[idx][mid].center_x
                        #y_long = m*x_short+b
                        y_long=m*x_short+b
                        if abs(y_short-y_long)<self.line_height:
                            #print "Short lines attached:",len(short_)
                            if index==-1 or dist>abs(y_short-y_long) or dist==abs(y_short-y_long) and short_left[idx][-1].center_x>center_x:
                                index=idx
                                dist=int(abs(y_short-y_long))
                                center_x=short_left[idx][-1]
                        idx+=1

                    if index!=-1:
                        belongs_to_line.append(short_left[index])
                        short_left.remove(short_left[index])

                        new_line=whole_long_line[gr_idx]
                        for n,belong in enumerate( belongs_to_line):
                            new_line=new_line+belong
                        m,b=self.get_parameters(new_line)

                    else:
                        changed=False

                #check if lines overlapp

                #
                short_array=self.check_overlap(belongs_to_line,line_groups[gr_idx][0])
                line_groups[gr_idx]=line_groups[gr_idx]+short_array
                #line_groups[gr_idx]=line_groups[gr_idx]+belongs_to_line

                line_groups[gr_idx].sort(lambda x,y: cmp(x[0].center_x,y[0].center_x))




            #check right side
            #---------------
            if short_right!=[]:
                belongs_to_line=[]
                m,b=self.get_parameters(whole_long_line[gr_idx])
                #m=parameter_long_line[gr_idx][-1][0]
                #b=parameter_long_line[gr_idx][-1][1]
                #print "+++++old m,b ",m,b
                changed=True
                while(changed==True):
                    idx=0
                    short_close_distance=[]
                    dist=image.height
                    index=-1
                    center_x=image.width
                    while idx< len(short_right):
                        mid=len(short_right[idx])/2
                        y_short=short_right[idx][mid].center_y
                        x_short=short_right[idx][mid].center_x
                        #y_long = m*x_short+b
                        y_long=m*x_short+b
                        if abs(y_short-y_long)<self.line_height:
                            if index==-1 or dist>abs(y_short-y_long) or dist==abs(y_short-y_long) and short_right[idx][-1].center_x<center_x:
                                index=idx
                                dist=int(abs(y_short-y_long))
                                center_x=short_right[idx][-1]
                        idx+=1

                    if index!=-1:
                        belongs_to_line.append(short_right[index])
                        short_right.remove(short_right[index])

                        new_line=whole_long_line[gr_idx]
                        for n,belong in enumerate( belongs_to_line):
                            new_line=new_line+belong
                        m,b=self.get_parameters(new_line)

                    else:
                        changed=False

                short_array=self.check_overlap(belongs_to_line,line_groups[gr_idx][-1])
                line_groups[gr_idx]=line_groups[gr_idx]+short_array
                #line_groups[gr_idx]=line_groups[gr_idx]+belongs_to_line

                line_groups[gr_idx].sort(lambda x,y: cmp(x[0].center_x,y[0].center_x))


            gr_idx+=1




        if self.clean==True:
            #CLEAN THE LINES        
            #try to find the rest of the lines, even if the height of the line is a little bit bigger, so that parts of some letters might be deleted
            for gr_number,group in enumerate(line_groups):
                to_append=[]
                for line_number,line in enumerate(group):
                    rect_to_append=[]
                    for rect_number,rect in enumerate(line):
                        if rect_number==len(line)-1:
                            break
                        #try to find the thicker not detected peaces of the line
                        if line_groups[gr_number][line_number][rect_number+1].center_x-rect.center_x>1:
                            #Look for Elements who fit in the holes

                            for runs in all_bigger_runs:
                                for run in runs:
                                    if (rect.center_x<run.center_x) and (run.center_x<line_groups[gr_number][line_number][rect_number+1].center_x) and ((abs(rect.center_y-run.center_y)<2*self.line_height) or (abs(line_groups[gr_number][line_number][rect_number].center_y-run.center_y)<2*self.line_height)) and (run.height<3*self.line_height):
                                        line_groups[gr_number][line_number].append(run)
                                        to_append.append(run)
                                        #rgb_image.draw_line(run.ul,run.ll,RGBPixel(255,255,255))
                    #try to find thicker not detected peaces between two lines
                    if line_number==len(group)-1:
                        break

                    for runs in all_bigger_runs:
                        for run in runs:
                            if (line[-1].center_x<run.center_x) and (run.center_x<group[line_number+1][0].center_x) and abs(run.center_x-group[line_number+1][0].center_x)<3*self.character_height and (group[line_number+1][0].center_x-line[-1].center_x)>0  and (group[line_number+1][0].center_x-line[-1].center_x)<2*self.character_height and len(line)>self.line_height and len(group[line_number+1])>self.line_height and ((abs(line[-1].center_y-run.center_y)<self.line_height*0.75) or (abs(group[line_number+1][0].center_y-run.center_y)<self.line_height*0.75)) and (run.height<3*self.line_height):
                                to_append.append(run)
                                #rgb_image.draw_line(run.ul,run.ll,RGBPixel(255,255,255))

                line_groups[gr_number].append(to_append)
                #line.sort(lambda a,b: cmp(a.center_x,b.center_x))


        if self.log_file:
            self.write_log_file(self.log_file)

        print "Found %d lines" % len(line_groups)

        for group in line_groups:
            for line in group:
                for rect in line:
                     self.image.draw_line(rect.ul,rect.ll,RGBPixel(255,255,255))

        return line_groups
                              

#-------------------------------------------------------------------------------
    def mark_lines(self):
        """Returns an RGB image with all found separator lines
highlighted in red.
"""
        rgb_image=self.original_image.to_rgb()
        for group in self.lines:
            #color=RGBPixel(random.randint(34,224),random.randint(34,224),random.randint(34,224))
            for line in group:
                for rect in line:
                    rgb_image.draw_line(rect.ul,rect.ll,RGBPixel(255,0,0))
                    #rgb_image.draw_line(rect.ul,rect.ll,color)

        return rgb_image

#-------------------------------------------------------------------------------
    def write_log_file(self, log_file):
        if log_file:
            f=open(logfile,"a")

            # falsepos = non removed pixel
            # falseneg = to many pixel removed

            if f.tell()==0:
                f.write("# detected lines,error , error , filename\n\n")

            iname=imagefile.split("/")
            error=0
            f.write("%d, %d,%d,  %s\n" %(len(self.lines),error,error,iname[-1]))

            f.close()

#---------------------------------------------------------------------

    def gui_show_lines(self):
        """Only for internal use in the GUI: display image with colored staves.
"""
        if has_gui.has_gui:
            self.GermanLutePage_sb = self.mark_lines()
            self.GermanLutePage_sb.display()
    		
#----------------------------------------------------------------------------
    def get_linear_equation(self):
        straight_line=[]
##         for line in len(self.lines):
##             straight_line.append([])

        for line in self.lines:
            full_line=[]
            for short_lines in line:
                full_line+=short_lines
            m,b=self.get_parameters(full_line)
            straight_line.append([m,b])

        return straight_line

#----------------------------------------------------------------------------
#----------------------------------------------------------------------------

###           VERTICAL LINES

#----------------------------------------------------------------------------
#----------------------------------------------------------------------------

    def find_bars(self):

        new_image=self.image.to_onebit()
        onebit_image=self.image.to_onebit()
        sec_onebit_image=self.image.to_onebit()

        onebit_image.filter_wide_runs(self.character_height,'black')

        ccs=onebit_image.cc_analysis()

        #delete all cc with a aspect ratio<3. So that there are only bars in the image. 
        for cc in ccs:
            if (cc.height<3) or float(cc.nrows/cc.ncols)<3.0:
                cc.fill_white()

        # get bar_width with most frequent_run
        bar_width = onebit_image.most_frequent_run('black','horizontal')


        sum_1=0.0
        sum_2=0.0
        N=0

        #get all runs in the image, for computing the standard deviation
        for x_iter in onebit_image.iterate_runs('black','horizontal'):
            for run in x_iter:
                sum_1+=run.width*run.width
                sum_2+=run.width
                #print run.width
                N+=1
        # evaluate the standard deviation
        sec_s_x=sqrt((N*sum_1-sum_2*sum_2)/(N*(N-1)))
        print "bar_width", bar_width

         #Get all runs and save those with a width <=bar_width
        #These are potential candidates for members of one line 
        x_list=[]
        all_bigger_runs=[]
        for x_iter in sec_onebit_image.iterate_runs('black','horizontal'):
            x_list.append([])
            all_bigger_runs.append([])
            for run in x_iter:
                    x_list[-1].append(run)

            #Searching for lines
        #-------------------------------------#
        #Go trough the list of potential candidates, and try to find related pixel lines 
        bars=[]
        longwidth=bar_width*2.5


        while True:
            x=0
            #jump over empty array elements
            while x<len(x_list) and len(x_list[x])==0:
                x+=1
            if x>=len(x_list):
                break

            #save all elements belonging to a line in run_list
            #first element of a line
            run_list=[x_list[x][0]]
            #delete already assigned elements
            del x_list[x][0]
            run_list_has_long_line=False
            x+=1
            jump=0
            #For each line breaks in range of 2*line_height allowed until the next part of the line is found. When after this distance no member of a line is found. The line is complete and the next line is searched.

            while x<len(x_list) and x<run_list[-1].center_y+self.line_height*1.5:
                found=False
                for run in x_list[x]+all_bigger_runs[x]:
                    #x ueberlappung
                    if max((run.ul_x,run_list[-1].ul_x))<min((run.lr_x,run_list[-1].lr_x)):
                        if  run.ul_y-run_list[-1].ul_y==1:
                            if found==False:
                                found=run
                            else:
                                if run.width>found.width:
                                    found=run
                if found!=False:
                    if found.width>longwidth:
                        if len(run_list)<self.character_height*1.5:
                            run_list_has_long_line=True
                        else:
                            break

                        if found not in all_bigger_runs[x]:
                            all_bigger_runs[x].append(found)
                    else:


                        run_list.append(found)
                    if found in x_list[x]:
                        x_list[x].remove(found)
                x+=1

            if not run_list_has_long_line:
                bars.append(run_list)
                for r in run_list:
                   # rgb_image.draw_line(r.ul,r.lr,RGBPixel(128,128,255))
                   pass


        ##sort bars into line_groups
        accords=[]
        short_accords=[]
        for k in range(len(self.lines)+1):
            accords.append([])
            short_accords.append([])

        bar_number=0
        while(bar_number<len(bars)) :
            if len(bars[bar_number])>0:
                for line_number,line in enumerate(self.lines):
                    if bars[bar_number][len(bars[bar_number])/2].center_y<(line[0][0].center_y):
                        accords[line_number].append(bars[bar_number])
                        bars.remove(bars[bar_number])
                        bar_number-=1
                        break
                else:
                    accords[len(accords)-1].append(bars[bar_number])
                    bars.remove(bars[bar_number])
                    bar_number-=1

            bar_number+=1

        #sort accords in line_groups
        accord_number=0
        while accord_number<len(accords):
            accords[accord_number].sort(lambda a,b: cmp(a[0].center_x,b[0].center_x))
            accord_number+=1


        #check if two bars overlapp, to put them in one bar_group
        accord_number=0
        changed=True
        while changed==True:
            changed=False
            while accord_number<len(accords):
                accord_idx=0
                while accord_idx<len(accords[accord_number])-1:
                    if len(accords[accord_number][accord_idx])<3:
                        accord_idx+=1
                        continue
                    sec_accord_idx=0
                    while sec_accord_idx< len(accords[accord_number])-1:
                        if len(accords[accord_number][sec_accord_idx])<3:
                            sec_accord_idx+=1
                            continue
                        if sec_accord_idx!=accord_idx:
                            #print "AAAAAAAAAAAAAAA" ,accord_number,len(accords[accord_number]),accord_idx,sec_accord_idx
                            top=max(accords[accord_number][accord_idx][0].center_y,accords[accord_number][sec_accord_idx][0].center_y)
                            down=min(accords[accord_number][accord_idx][-1].center_y,accords[accord_number][sec_accord_idx][-1].center_y)
                            overlapped=down-top

                            if accords[accord_number][accord_idx][0].center_y > accords[accord_number][sec_accord_idx][0].center_y and abs(accords[accord_number][accord_idx][0].center_x- accords[accord_number][sec_accord_idx][-1].center_x)<self.line_height*2.0 and overlapped <self.line_height and abs(accords[accord_number][accord_idx][0].center_y - accords[accord_number][sec_accord_idx][-1].center_y)<self.line_height *2.0:

                                    accords[accord_number][accord_idx]+=accords[accord_number][sec_accord_idx]
                                    accords[accord_number][accord_idx].sort(lambda a,b:cmp(a.center_y,b.center_y))
                                    accords[accord_number].remove(accords[accord_number][sec_accord_idx])
                                    changed=True
                                    if len(accords[accord_number][sec_accord_idx])>3:
                                        accord_idx=0
                                    break
                            elif  accords[accord_number][accord_idx][0].center_y <= accords[accord_number][sec_accord_idx][0].center_y and abs(accords[accord_number][accord_idx][-1].center_x- accords[accord_number][sec_accord_idx][0].center_x)<self.line_height*2.0  and overlapped<self.line_height  and abs(accords[accord_number][accord_idx][-1].center_y - accords[accord_number][sec_accord_idx][-1].center_y)<self.line_height *2.0 :
                                    accords[accord_number][accord_idx]+=accords[accord_number][sec_accord_idx]
                                    accords[accord_number][accord_idx].sort(lambda a,b:cmp(a.center_y,b.center_y))
                                    accords[accord_number].remove(accords[accord_number][sec_accord_idx])
                                    changed=True
                                    if len(accords[accord_number][sec_accord_idx])>3:
                                        accord_idx=0
                                    break

                        sec_accord_idx+=1
                    accord_idx+=1
                accord_number+=1



        #check if rects in a bar overlapp each other...
        #if they overlapp delete the thinner one
        accord_number=0
        while(accord_number<len(accords)):
            accord_idx=0
            while(accord_idx< len(accords[accord_number])):
                rect_idx=0
                while(rect_idx< len(accords[accord_number][accord_idx])-1):
                    sec_rect_idx=0

                    while(sec_rect_idx< len(accords[accord_number][accord_idx])):
                        if rect_idx!=sec_rect_idx:
                            if accords[accord_number][accord_idx][rect_idx].center_y==accords[accord_number][accord_idx][sec_rect_idx].center_y:
                                if accords[accord_number][accord_idx][rect_idx].width>accords[accord_number][accord_idx][sec_rect_idx].width:
                                    accords[accord_number][accord_idx].remove(accords[accord_number][accord_idx][sec_rect_idx])
                                    rect_idx=-1
                                    break
                                else:
                                    accords[accord_number][accord_idx].remove(accords[accord_number][accord_idx][rect_idx])
                                    rect_idx=-1
                                    break

                        sec_rect_idx+=1
                    rect_idx+=1
                accord_idx+=1
            accord_number+=1


        #delete to short or to big lines
        # short_line: smaller than 2.5*character_height,
        #             smaller than distance_between_two_lines/3
        #big_line: higer than 1.5 * distance_between_two_lines
        #these lines are because of their height, no bar candidates
        accord_number=0
        while(accord_number<len(accords)):
            accord_idx=0
            while(accord_idx< len(accords[accord_number])):
                if len(accords[accord_number][accord_idx])<self.distance_between_two_lines/3.0 or len(accords[accord_number][accord_idx])<self.character_height*2.5 or len(accords[accord_number][accord_idx])>1.5 *self.distance_between_two_lines:
                    for rect in accords[accord_number][accord_idx]:
                            #rgb_image.draw_line(rect.ul,rect.lr,RGBPixel(0,255,128))
                            pass
                    accords[accord_number].remove(accords[accord_number][accord_idx])
                    accord_idx-=1
                accord_idx+=1
            accord_number+=1


        for acc in accords:
            acc.sort(lambda a,b:cmp(a[0].center_x,b[0].center_x))

        ## get med_values of one line
        ## to filter all as bar detected flags from the original bars, get the med_values of one line.
        ## if there is a  to big distance differences between the bars, one line is not detected, and the lines are not deleted because of y-distance is to big. 

        acc_height=[]
        acc_y_end=[]
        error=[]
        for a_number,acc in enumerate(accords):
            acc_height.append([])
            acc_y_end.append([])
            error.append([])
            for a in acc:
                if len(a)==0:
                    continue
    ##             if len(acc)==0 or len(acc[-1])==0 or len(acc[0])==0:
    ##                 continue
                if acc_height[a_number]==[]:

                    acc_height[a_number]=(a[-1].lr_y-a[0].ul_y)
                    acc_y_end[a_number]=(a[-1].ll_y)
                else:

                    acc_height[a_number]=(acc_height[a_number]+a[-1].ll_y-a[0].ul_y)/2
                    if abs( acc_y_end[a_number]-a[-1].ll_y)>self.distance_between_two_lines:
                        if error[a_number]==[]:
                            error[a_number]=1
                        else:
                            error[a_number]+=1
                    else:
                        acc_y_end[a_number]=(acc_y_end[a_number]+a[-1].ll_y)/2
        #print error
        accord_number=0
        while accord_number<len(accords):
            accord_idx=0
            while accord_idx<len(accords[accord_number])-1:
                if len(accords[accord_number][accord_idx])==0:
                    continue

     ##            if len(accords[accord_number][-1])==0 or len(accords[accord_number][0])==0:
    ##                 continue
                dist=accords[accord_number][accord_idx][-1].ll_y
                if abs(accords[accord_number][accord_idx][-1].ll_y-acc_y_end[accord_number])>2.0*self.character_height and error[a_number]<3 :
                    #for rect in accords[accord_number][accord_idx]:
                            #rgb_image.draw_line(rect.ul,rect.lr,RGBPixel(0,255,255))
                            #pass
                    accords[accord_number].remove(accords[accord_number][accord_idx])
                    accord_idx=-1

                elif (accords[accord_number][accord_idx][-1].ll_y-accords[accord_number][accord_idx][0].ul_y)<acc_height[accord_number]/2:
                    #for rect in accords[accord_number][accord_idx]:
                            #rgb_image.draw_line(rect.ul,rect.lr,RGBPixel(255,255,0))
                            #pass
                    accords[accord_number].remove(accords[accord_number][accord_idx])
                    accord_idx=-1


                accord_idx+=1
            accord_number+=1



        accord_number=0
        while accord_number<len(accords):
            accord_idx=0
            while accord_idx<len(accords[accord_number])-1:
                if abs(accords[accord_number][accord_idx][-1].ll_y-accords[accord_number][accord_idx+1][-1].ll_y)>self.character_height and  abs(accords[accord_number][accord_idx][-1].ll_y-accords[accord_number][accord_idx+1][-1].ll_y)<self.distance_between_two_lines:
                    if len(accords[accord_number][accord_idx])>len(accords[accord_number][accord_idx+1]) and accord_idx!=0:
                       ##  for rect in accords[accord_number][accord_idx+1]:
    ##                         rgb_image.draw_line(rect.ul,rect.lr,RGBPixel(128,0,128))
    ##                         #pass
                        accords[accord_number].remove(accords[accord_number][accord_idx+1])
                        accord_idx=-1
                    elif len(accords[accord_number][accord_idx])<len(accords[accord_number][accord_idx+1]) and accord_idx+1!=len(accords[accord_number])-1 :

    ##                     for rect in accords[accord_number][accord_idx]:
    ##                         rgb_image.draw_line(rect.ul,rect.lr,RGBPixel(128,0,128))
    ##                         #pass
                        accords[accord_number].remove(accords[accord_number][accord_idx])
                        accord_idx=-1


                accord_idx+=1
            accord_number+=1

        acc_number=1
        for accord in accords:
            for acc in accord:
                acc_number+=1
                for rect in acc:
                    self.image.draw_line(rect.ul,rect.lr,RGBPixel(255,255,255))


        return accords


    #----------------------------------------------------------
    def mark_bars(self):
        """Returns an RGB image with all heuristically found barlines
highlighted in red. Note that this only makes sense when you have initially
chosen the heuristic bar detection in the constructor.
"""
##        if self.bar_detection:
##             bar_image=self.image.to_rgb()
##             for bar in bar_ccs:
##                 bar_image.highlight(bar,RGBPixel(0,255,0))

##             return bar_image
                

        
        if self.bar_detection:
            bar_image=self.original_image.to_rgb()
            acc_number=1
            for accord in self.bars:
                for acc in accord:
                    acc_number+=1
                    for rect in acc:
                        bar_image.draw_line(rect.ul,rect.lr,RGBPixel(255,0,0))

            return bar_image
        else:
            print "Error: bar_detection==False"
            return self.image



    #---------------------------------------------------------------------

    def gui_show_bars(self):
        """Only for internal use in the GUI: display image with colored staves.
"""
        if has_gui.has_gui:
            self.GermanLutePage_sb = self.mark_bars()
            self.GermanLutePage_sb.display()


    #----------------------------------------------------------------------
    def get_bar_ccs(self):
        bars=[]

        for bar_line in self.bars:
           for rect_bar in bar_line:
               bars.append(union_images(rect_bar))

        return bars
    
            
        
