# -*- mode: python; indent-tabs-mode: nil; tab-width: 4 -*-
# vim: set tabstop=4 shiftwidth=4 expandtab:

import image_utilities
from to_abcmusic import *
import sys,math

symbol_table=["fret.a", "fret.b", "fret.c",
              "fret.d", "fret.e", "fret.f",
              "fret.g", "fret.h", "fret.i",
              "fret.k", "fret.l", "fret.m",
              "_group.fret.a", "_group.fret.b", "_group.fret.c",
              "_group.fret.d", "_group.fret.e", "_group.fret.f",
              "_group.fret.g", "_group.fret.h", "_group.fret.i",
              "_group.fret.k", "_group.fret.l", "_group.fret.m",
              "flag.1", "flag.2", "flag.4",
              "flag.8", "flag.16", "flag.32",
              "flag.64", "flag.128",
              "_group.flag.1", "_group.flag.2", "_group.flag.4",
              "_group.flag.8", "_group.flag.16", "_group.flag.32",
              "_group.flag.64", "_group.flag.128",
              "bar", "_group.bar",
              "dot",
              "deco.stroke", "deco.v_line",
              "deco.hash.here", "deco.star.here", "_group.deco.hash.here",
              "fermata", "_group.fermata",
              "time.c", "time.c_cut", "time.3",
              "trash"]

#debugout = ["after singleglyph", "singleglyphs", "check_flaggroup", 
#            "before grouping", "after grid grouping", "doublette"]
debugout = []

def debug_print_glyphs(glyphs, prefix="DBG"):
    if prefix in debugout: # ["after singleglyph","lastx==g.x"]: #"2"]:
        print '\033[1;32m[DBG Glyphs %s]\033[1;m' % prefix
        for g in glyphs:
            print "     ", prefix, g.id_name, g.x_number, g.gridnr, g

def debug_print(a, prefix="DBG"):
    if prefix in debugout:
        print '\033[1;31m[DBG %s]\033[1;m' % prefix, a

##############################################################################
class TabGlyphList:
##############################################################################
    """Class for storing a list of image objects.

An object of the class *TabGlyphList* provides the following functionality:

 - Storing all connected components (glyphs) of a tablature in a list
 - Providing information about the position (staff, course) of each glyph
 - Sorting glyphs by their positions (staff, course, chord)
 - Output of all glyphs in a specified format

The constructor_ of a *TabGlyphList* object adds several public properties
to each image of its image list. Those properties can be used to determine
the logical position of each symbol in the tablature.

:Author: Christoph Dalitz and Thomas Karsten
"""

    ######################################################################
    # constructor
    #
    def __init__(self, imgs):
        """.. _constructor:

Creates and returns a *TabGlyphList* object.

Signature:

  ``__init__(imgs)``

with

  *imgs*:
    List of images (connected components or glyphs) that is referenced by the
    *TabGlyphList* object. The images have to be Onebit-images.

A *TabGlyphList* contains the public properties as described below:

+----------------------+-------------------------------------------------+
| ImageList ``glyphs`` | Reference to the list of images as given in the |
|                      | constructor                                     |
+----------------------+-------------------------------------------------+

Additionally, new public properties are introduced in the constructor of this
object for each item in ``glyphs``:

+------------------+---------------------------------------------------------+
| Int ``staff``    | Number of staff to which the glyph belongs (from top    |
|                  | to bottom, starting with 1)                             |
+------------------+---------------------------------------------------------+
| Int ``course``   | Number of the course on which the glyph is located      |
|                  | (from top to bottom, starting with 1 in its staff).     |
|                  | Glyphs above the first course get the course number 0   |
|                  | (e.g. flags).                                           |
+------------------+---------------------------------------------------------+
| Int ``x_number`` | Position of the glyph in its staff. ``x_number`` starts |
|                  | with 1 and is the same for all glyphs of a vertically   |
|                  | aligned chord. The value of ``x_number`` is incremented |
|                  | when detecting a new glyph that does not belong to the  |
|                  | current chord.                                          |
+------------------+---------------------------------------------------------+
| Int ``middle_x`` | x-position of the center of the glyph.                  |
+------------------+---------------------------------------------------------+
| Int ``middle_y`` | y-position of the center of the glyph.                  |
+------------------+---------------------------------------------------------+
| Int ``gridnr``   | position in grid:                                       |
|                  |  ``0`` for glyph not within grid  or flag,              |
|                  |  ``n`` for glyph in n-th chord under grid (starts at 1),|
|                  |  ``-n`` for grid/flag including n chords.               |
+------------------+---------------------------------------------------------+

The properties ``offset_x`` and ``offset_y`` in Gamera images refer to the
upper left corner. Thus the area middle point is (``offset_x`` + ``ncols`` / 2,
``offset_y`` + ``nrows`` / 2). Although it was possible to use the center of
gravity (also known as first moment) instead, its computation was more time
consuming. A *TabGlyphList* object uses the middle point of the area of a
glyph to determine its position.

The initial value of ``staff``, ``course`` and ``x_number`` is set to -1.
These values are determined and set properly in the public method
set_glyph_properties_.
"""
        self.glyphs=imgs         # list of TabGlyphs
        self.glyph_props_set = 0 # for efficiency to avoid multiple calls

        # initialize all glyphs with default values
        for gl in self.glyphs:
            gl.staff=-1
            gl.course=-1
            gl.x_number=-1
            gl.middle_x=int(gl.offset_x+gl.ncols*0.5)
            gl.middle_y=int(gl.offset_y+gl.nrows*0.5)
            gl.gridnr=-1

        self.overlapthreshold = 0.3# 0.23
        self.overlapthreshold_flaggroup = 0.23# 0.23

    ######################################################################
    # set_glyph_properties
    #
    # gets the tablature position of each glyph in the list and sets
    # their properties staff, course and x_number. the list of glyphs
    # will be sorted in this method.
    #
    # parameters: tab: initialized TabPage object
    #
    # toom, 2004-06-01
    #
    def set_glyph_properties(self, tab):
        """Sets important properties of each glyph to determine its
position.


Signature:

  ``set_glyph_properties(tab)``

with

  *tab*:
    Initialized TabPage_ object where its method `remove_staves`_ has to be
    called before.

.. _TabPage: gamera.toolkits.otr.otr_staff.TabPage.html
.. _`remove_staves`: gamera.toolkits.otr.otr_staff.TabPage.html#remove-staves

This method gets the tablature position of each glyph in ``self.glyphs`` and
sets its properties ``staff``, ``x_number`` and ``course``. These properties
describe the position of each glyph in the tablature (see the image below and
the constructor_).


.. image:: images/pagelayout.png

After the method has finished, the glyphs of the *TabGlyphList* are sorted
by ``staff``, ``x_number``, ``course``, and ``offset_x`` (the latter is for
symbols that are at the same position, e.g. flags with dots.

.. note:: Glyphs without sensible staff or course obtain the value ``-1`` for
          ``staff`` or ``course``.
"""

        #
        # sets the x_number for the glyph 'glyphs[index]' by comparing
        # its position with the position of all glyphs until
        # 'glyphs[index-max_glyphs]'.
        # in case the current glyph could be "added" to a chord, the
        # return value is 1, otherwise 0.
        #
        # variables taken from the calling function:
        #    glyphs          glyph list
        #    index           index of current glyph
        #    max_glyphs      number of previous glyphs to test
        #    cur_x_number    current x_number (position)
        #
        def get_x_number():
            cnt=0
            g=self.glyphs[index]    # current glyph
            cmp_gl=index-1      # previous glyph
            found=0

            # symbols ranging from top to bottom of a staff get
            # an own x_number
            if g.staff >= 1 and\
                    g.nrows > tab.staffspace_height\
                    *len(tab.stafflist[g.staff-1].yposlist)\
                    *0.85:
                g.x_number=cur_x_number+1
                return 0

            while cmp_gl >= 0 and cnt < max_glyphs and not found:

                if g.staff != self.glyphs[cmp_gl].staff:
                    break

                # ignore wide glyphs
                if self.glyphs[cmp_gl].ncols > tab.staffspace_height*1.6 or \
                        self.glyphs[cmp_gl].match_id_name("*flag*") or \
                        self.glyphs[cmp_gl].match_id_name("*bar*"):

                    if False:
                       print "ignoring for comparison"
                    cmp_gl-=1
                    continue

                overlap=self.__get_max_overlap(g, \
                        self.glyphs[cmp_gl])

                # the current glyph gets the x_number of the
                # glyph in its column
                if overlap > self.overlapthreshold:
                    #check for existing glyph at this position
                    g.x_number=self.glyphs[cmp_gl].x_number
                    found=True

                # test next glyph
                cmp_gl-=1
                cnt+=1

            if not found:
                g.x_number=cur_x_number+1
                found=0

            return found

        # remember that we have called this function
        self.glyph_props_set = 1

        # how many glyphs should the current glyph be compared with?
        # (used for determining the exact x_number)
        max_glyphs=6

        # set properties of each glyph
        for index, g in enumerate(self.glyphs):

            try:
                g.middle_x

            except AttributeError:
                g.middle_x=int(g.offset_x+g.ncols*0.5)
                g.middle_y=int(g.offset_y+g.nrows*0.5)

            tmp=tab.get_staffcourse(g.middle_x, g.middle_y)

            # position of glyph could be determined
            if len(tmp) == 2:
                g.staff=tmp[0]
                g.course=tmp[1]

            # position of glyph is not sure
            else:
                g.staff=-1
                g.course=-1

            g.x_number=-1

            # calculate length if flag
            g.gridnr = 0
            if g.match_id_name("flag*"):
                if False:
                    print "preprocessing beam"
                lengths = g.id_name[0][1].split('.')[1:]
                if False:
                    print lengths
                if len(lengths) == 1:
                    g.gridnr = -1

                else:
                    for i in range(1, len(lengths), 2):
                        g.gridnr -= int(lengths[i])

                if False:
                    print "length: %d" % g.gridnr
            elif g.match_id_name("fret.*"):
                fret = g.id_name[0][1].split('.')
                if len(fret[1]) > 1:
                    try:
                        g.course = int(fret[1][1:])
                    except:
                        print "Course error"
                g.classify_heuristic("%s.%s" % (fret[0], fret[1][0]))




        # sort by pixel values
        self.glyphs.sort(self.cmp_glyph_coords())

        #
        # the glyphs are now sorted by their pixel positions, but this
        # is not exact enough. in some situations glyphs that are
        # horizontally very near to each other will be sorted as if
        # they belonged to 1 chord. so the list is not really sorted
        # by chords (there are still some mismatches).
        # to be more accurate, we set the x_number (x-position in
        # their staff) of a glyph by comparing its position with the
        # position of previous glyphs. if it overlaps with a previous
        # glyph, it will get its x_number, otherwise a new chord is
        # 'created' (by incrementing the x_number).
        #
        cur_staff=-1
        cur_x_number=1

        for index, g in enumerate(self.glyphs):
            #skip flags for later grouping
            if g.match_id_name("*flag*"):
                continue
            # starting a new staff
            if cur_staff != g.staff:
                cur_staff=g.staff
                cur_x_number=1
                g.x_number=1

            # decide about the x_number of the glyph
            else:
                found=get_x_number()

                if not found:
                    cur_x_number+=1

        # sort the list again, but use staff, course and x_number
        # instead of pixel values (middle_x, middle_y)
        self.glyphs.sort(self.cmp_glyph_properties_grid)

        debug_print_glyphs(self.glyphs, "before grouping")

        if self.__correct_doublette_positions(tab) > 0:
            print "SORTING"
            self.glyphs.sort(self.cmp_glyph_properties)

        debug_print_glyphs(self.glyphs, "before grouping")
        modified=self.__group_grid(tab)
        debug_print_glyphs(self.glyphs, "after grid grouping")
        modified &= self.__group_deco(tab)
        debug_print_glyphs(self.glyphs, "after deco grouping")
        self.barline_correction(tab)
        debug_print_glyphs(self.glyphs, "after barline correction")
        self.special_chords()

        self.glyphs = [g for g in self.glyphs if g.x_number >= 0]
        self.glyphs.sort(self.cmp_glyph_properties)



    def __correct_doublette_positions(self, tab):
        """corrects glyphs which are put on same position"""
        exclude_from_splitting = ["*deco.*.left", "*deco.*.right"]
        curx_glyphs = []
        coursehistogram = [0]*7
        curx = -2
        xoffset = 0
        def split_x_number():
            def clusteroverlap(a,b):
                maxoverlap = 0
                for aa in a:
                    for bb in b:
                        if aa != bb:
                            overlap = self.__get_max_overlap(aa,bb)
                            if overlap > maxoverlap:
                                maxoverlap = overlap
                return maxoverlap


            groups = [[x] for x in curx_glyphs]
            maxoverlap = 1000
            while len(groups) > 2 and maxoverlap > 0:
                maxx = None
                maxy = None
                maxoverlap = -1
                for x in groups:
                    for y in groups:
                        if x != y:
                            overlap = clusteroverlap(x,y)
                            if overlap > maxoverlap:
                                maxoverlap = overlap
                                maxx = x
                                maxy = y
                if maxx != None:
                    maxx.extend(maxy)
                    groups.remove(maxy)

            if maxoverlap == 0:
                return 0
            mean0 = sum([z.middle_x for z in groups[0]]) / len(groups[0])
            mean1 = sum([z.middle_x for z in groups[1]]) / len(groups[1])

            if mean1 > mean0:
                for z in groups[1]:
                    z.x_number += 1
            else:
                for z in groups[0]:
                    z.x_number += 1

            return 1


        curstaff = -5
        for index, g in enumerate(self.glyphs):
            if g.match_id_name("*flag*"):
                continue

            if g.x_number != curx:
                debug_print_glyphs(curx_glyphs, "doublette")

                if max(coursehistogram) > 1 and curx >= 0:
                    v = split_x_number()
                    curx = g.x_number+v
                    if v > 0:
                        for x in (self.glyphs[index:]):
                            if x.staff != curstaff:
                                break

                            x.x_number += 1

                else:
                    curx = g.x_number
                curstaff = g.staff


                debug_print_glyphs(curx_glyphs, "doublette")
                curx_glyphs = []
                coursehistogram = [0]*10

            curx_glyphs.append(g)
            skipglyph = False #skip special glyphs in histogram
            for name in exclude_from_splitting:
                if g.match_id_name(name):
                    skipglyph = True


            if g.course >= 0 and g.course < 10 and not skipglyph:
                coursehistogram[g.course] += 1

        xoffset = 0

        return xoffset




    ##########################################################################
    #
    # 2012-07-16 Christian Brandt
    def __group_grid(self, tab, fast=True):
        modified=0  # indicates modification
        update=0    # return value
        rm_glyphs=[]    # glyphs that should be removed

        # maximum possible distance between dot and flag to
        # accept a grouping
        max_distance=tab.staffspace_height
        gridcount = 1 #0
        lastx = -2
        curflag = None

        def get_flagcourse():
            #gets major course containing flags and creates a map of flags sorted by staff
            courses = {}
            flagmap = {}
            lastflagidx=-1
            for g in self.glyphs:
                if g.match_id_name("*flag*"):
                    lastflagidx += 1
                    if g.staff >= 0:
                        if courses.has_key(g.course):
                            courses[g.course]+=1
                        else:
                            courses[g.course] = 1

                        if not flagmap.has_key(g.staff):
                            flagmap[g.staff] = [g]
                        else:
                            flagmap[g.staff].append(g)

            count = 0
            course = -1
            for c in courses:
                if courses[c] > count:
                    count = courses[c]
                    course = c

            return course,flagmap


        flagcourse,flags_by_staff = get_flagcourse()
        debug_print("flagcourse is %d" % flagcourse, "flagcourse")


        ###############################################################################
        def check_flagdot(g, cmp_gl, curflag, gridcount):
            # this method checks whether g is a dot belonging to cmp_gl in case
            # cmp_gl is a flag
            found = 0
            if cmp_gl.match_id_name("*flag*") and cmp_gl.staff == g.staff \
                    and g.course == flagcourse and g.match_id_name("*dot*"):

                debug_print("found flag candidate", "check_flagdot")
                left_g = g.offset_x
                left_cmp = cmp_gl.offset_x
                right_g = g.offset_x + g.ncols
                right_cmp = cmp_gl.offset_x + cmp_gl.ncols
                if left_g >= left_cmp and right_g <= right_cmp:
                    factor = ((left_g+right_g)/2.0-left_cmp) / (cmp_gl.ncols)
                    debug_print(("dotflag found", cmp_gl.gridnr, g.gridnr, factor), "check_flagdot")
                    g.gridnr = int(factor * (-cmp_gl.gridnr)) + 1
                    debug_print(("newgridnr", g.gridnr), "check_flagdot")
#                            g.x_number = cmp_gl.x_number
                    found = 1
                    "print voerlap"
                #nearby
                elif left_g > right_cmp and left_g-right_cmp < (right_g-left_g) * 4 and g.x_number <= cmp_gl.x_number+1:
                    g.gridnr = 1
                    found = 1
                    debug_print(("nearby"), "check_flagdot")
                else:
                    debug_print(("not found"), "check_flagdot")
                if found:
                    if cmp_gl != curflag:
                        gridcount = 0
                        curflag = cmp_gl

                    if cmp_gl.x_number < 0:
                        cmp_gl.x_number = g.x_number - g.gridnr + 1
                    else:
                        g.x_number = cmp_gl.x_number

            return found,curflag,gridcount


        ###############################################################################
        def check_flaggroup(g, cmp_gl,curflag,gridcount):
            #checks whether g belongs to grid cmp_gl and sets properties of g
            found = 0
            if cmp_gl.match_id_name("*flag*") and cmp_gl.staff == g.staff and \
                    cmp_gl.course == flagcourse and not g.match_id_name("time.*"): # and not g.match_id_name("*dot*"):

                overlap = self.__get_overlap(g, cmp_gl)
                if overlap < self.overlapthreshold_flaggroup:
                    #chek for nearby
                    debug_print(("F", g, cmp_gl, g.id_name, cmp_gl.id_name), "check_flaggroup")

                    g.gridnr = 0
                    return found,curflag,gridcount

                debug_print(("R", g, cmp_gl, g.id_name, cmp_gl.id_name), "check_flaggroup")

                if cmp_gl != curflag:
                    curflag = cmp_gl
                    gridcount = 1


                found = 1

                #gridcount += 1
                debug_print(("xnu", g.x_number, cmp_gl.x_number),"check_flaggroup")
                if cmp_gl.x_number < 0:
                    cmp_gl.x_number = g.x_number
                else:
                    cmp_gl.x_number = min(g.x_number, cmp_gl.x_number)
                #cmp_gl.x_number = min(g.x_number,max(0,cmp_gl.x_number))
                if gridcount > (-cmp_gl.gridnr):
                    debug_print(("full"), "check_flaggroup")
                    g.gridnr = 0
                else:
                    debug_print(("not full"), "check_flaggroup")
                    g.gridnr = gridcount
                    found=1

            return found,curflag,gridcount



        rm_glyphs = []

        flagsperxnumber = [ {}] * (len(tab.get_staffboxes())+1)
        for index, g in enumerate(self.glyphs):
            if False:
                print "xpos:",g.x_number, "class:",g.id_name, "gridnr:", \
                    g.gridnr, "gridcount:", gridcount
            if g.match_id_name("*flag*"):
                continue
            if g.match_id_name("*bar*"):
                g.gridnr = 0
                lastx=g.x_number ##TODO???
                continue
            if lastx == g.x_number:
                debug_print_glyphs([g], "lastx==g.x")
                g.gridnr = gridcount
            else:
                gridcount += 1

            #only search in flags in current staff for speedup
            try:
                cmpflags = flags_by_staff[g.staff]
            except:
                #alternative solution could be to fall back to all glyphs if no flags in current staff
                #cmpflags = self.glyphs
                cmpflags = []

            try:
                nflag = flagsperxnumber[g.staff][g.x_number]
                g.gridnr = nflag[1] #g.x_number - nflag.x_number + 1
            except:
                pass

            lastx = g.x_number
            cmp_index_skip = 1
            maxcheck = 1000
            cmp_index = 0#max(0, index-maxcheck)
            stop = min(len(cmpflags), maxcheck)
            found = 0

            debug_print_glyphs([g], "2_1")
            while cmp_index >= 0 and cmp_index < stop and found == 0:
                cmp_gl = cmpflags[cmp_index]
                debug_print_glyphs([cmp_gl], "2")
                found, curflag, gridcount = check_flagdot(g, cmp_gl, curflag, gridcount)

                if not found:
                    found, curflag, gridcount = check_flaggroup(g, cmp_gl,curflag,gridcount)

                if not found:
                    cmp_index += cmp_index_skip
                else:
                    if not flagsperxnumber[g.staff].has_key(g.x_number):
                        flagsperxnumber[g.staff][g.x_number] = (cmp_gl, g.gridnr)

            if not found:
                gridcount = 0

        debug_print_glyphs(self.glyphs, "before remove")
        debug_print_glyphs(rm_glyphs, "to remove")


        for g in rm_glyphs:
            self.glyphs.remove(g)


        def count_group_size(glyphs,cmp_index):
            maxpos = 0
            lastx = glyphs[cmp_index].x_number
            current_staff = glyphs[cmp_index].x_number
            flagsize = 0
            while cmp_index < len(glyphs):
                cmp_gl = glyphs[cmp_index]
                if cmp_gl.gridnr > maxpos:
                    maxpos = cmp_gl.gridnr
                elif lastx < cmp_gl.x_number or cmp_gl.staff > current_staff:
                    break
                elif flagsize == 0:
                    flagsize = -cmp_gl.gridnr
                cmp_index += 1

            return (maxpos, flagsize, flagsize-maxpos)


        debug_print_glyphs(self.glyphs, "b1")
        self.glyphs.sort(self.cmp_glyph_properties_grid)
        debug_print_glyphs(self.glyphs, "b2")


        #look for ungrouped chords and non-empty glyphs
        nonempty = {}
        singleglyphs = []
        for index, g in enumerate(self.glyphs):
            if g.gridnr < 0:
                size = count_group_size(self.glyphs, index)
         
                if size[2] > 0:
                    if not nonempty.has_key(g.staff):
                        nonempty[g.staff] = {}

                    nonempty[g.staff][g.x_number] = (index, g, size[2])
          
            elif g.gridnr == 0 and not g.match_id_name("*bar*") and not g.match_id_name("time*"):
                debug_print(("ungrouped:",g,"class:",g.id_name,"staff:",g.staff,"xpos:",g.x_number), "ungrouped")
                singleglyphs.append((g,index))
            else:
                debug_print(("grouped", g, "class:", g.id_name, "staff:", g.staff, "xpos", g.x_number, "gridnr", g.gridnr), "grouped")

        debug_print(nonempty, "nonempty")

        # attach remaing glyphs to nonempty glyphs
        lastcorrected = -2
        cur_staff = -1
        for (g,index) in singleglyphs:
            #try right
            attached = False
            if g.x_number == lastcorrected:
                g.gridnr = 0.5
            debug_print((g.get_main_id(), g.x_number, g.gridnr), "singleglyphs")
            if nonempty.has_key(g.staff):
                #leftdistance
                leftdistance = -1
                rightdistance = -1
                if nonempty[g.staff].has_key(g.x_number):
                    flag_idx, flag_g, flag_remain = nonempty[g.staff][g.x_number]
                    if flag_remain > 0:
                        leftdistance = abs(flag_g.middle_x - g.middle_x)

                if nonempty[g.staff].has_key(g.x_number+1):
                    flag_idx, flag_g, flag_remain = nonempty[g.staff][g.x_number+1]
                    if flag_remain > 0:
                        rightdistance = abs(flag_g.middle_x - g.middle_x)

                if leftdistance < rightdistance and leftdistance > 0:
                    debug_print(("left"), "singleglyphs")
                    flag_idx, flag_g, flag_remain = nonempty[g.staff][g.x_number]
                    nonempty[g.staff][g.x_number] = (flag_idx, flag_g, flag_remain-1)

                    g.gridnr = -0.5
                    flag_g.x_number = g.x_number
                    lastcorrected = g.x_number
                    nonempty[g.staff][g.x_number] = (index, flag_g, flag_remain-1)
                elif rightdistance > 0:
                    debug_print(("right"), "singleglyphs")
                    flag_idx, flag_g, flag_remain = nonempty[g.staff][g.x_number+1]
                    del nonempty[g.staff][g.x_number+1]
                    tmp = self.glyphs[flag_idx]
                    self.glyphs[flag_idx] = g
                    self.glyphs[index] = tmp
                    g.gridnr = 0.5
                    flag_g.x_number = g.x_number
                    lastcorrected = g.x_number
                    nonempty[g.staff][g.x_number] = (index, flag_g, flag_remain-1)
                else:
                    debug_print(("not attached"), "singleglyphs")






        debug_print_glyphs(self.glyphs, "after singleglyph")

        inserted = 0
        for index,g in enumerate(self.glyphs):
            if g.gridnr == 0.5:
                g.gridnr = 1
                inserted += 1
            elif g.gridnr == -0.5:
                g.gridnr = -100
                inserted += 1
            elif inserted:
                if g.gridnr > 0:
                    g.gridnr += inserted
                else:
                    inserted -= inserted
            debug_print((g, g.x_number, g.id_name, g.gridnr), "inserted")

        debug_print_glyphs(self.glyphs, "after singleglyph")



        return update



    ##########################################################################
    #
    # 2012-06-01 Christian Brandt
    def __group_deco(self, tab):
        modified=0  # indicates modification
        update=0    # return value
        rm_glyphs=[]    # glyphs that should be removed

        # maximum possible distance between dot and flag to
        # accept a grouping
        max_distance=tab.staffspace_height

        for index, g in enumerate(self.glyphs):
            cmp_index =index
            cmp_index_skip = 0
            if(g.match_id_name("deco.*.left")):
                cmp_index_skip = -1
            elif(g.match_id_name("deco.*.right")):
                cmp_index_skip = 1

            if cmp_index_skip != 0:
                cur_x_number=g.x_number
                x_to_check = cur_x_number+cmp_index_skip
                prev_x = min(x_to_check, cur_x_number)
                next_x = max(x_to_check, cur_x_number)

                found=0
                cmp_gl=self.glyphs[cmp_index]

                while cmp_index >= 0 and cmp_index < len(self.glyphs) and\
                        cmp_gl.x_number >= prev_x and\
                        cmp_gl.x_number <= next_x\
                        and found == 0:

                    if cmp_gl.match_id_name("*fret*") and cmp_gl.staff == g.staff and cmp_gl.course == g.course:
                        found = 1
                        g.x_number = cmp_gl.x_number
                        g.gridnr = cmp_gl.gridnr
                        update = 1
                    else:
                        cmp_index += cmp_index_skip
                        cmp_gl = self.glyphs[cmp_index]

        return update




    ######################################################################
    # Bool __group_flag_and_dot
    #
    # problem: when there is a flag with a dot (e. g. indicating a
    #          3/4 flag) which do not overlap, then our algorithm in
    #          'set_glyph_properties' sets different x_numbers. due to
    #          this, all following glyphs overlapping with the dot, will
    #          get the x_number of the dot. then the flag is 'left alone'
    #          in its chord.
    #
    # this function solves the problem as described above. this function
    # only works *after* classifying glyphs (because it needs to know
    # about 'dot'). the return value indicates whether the list of glyphs
    # was modified or not. if it was changed, it should be sorted after
    # calling this function.
    #
    # toom, 2004-09-08
    #
    def __group_flag_and_dot(self, tab):

        modified=0  # indicates modification
        update=0    # return value
        rm_glyphs=[]    # glyphs that should be removed

        # maximum possible distance between dot and flag to
        # accept a grouping
        max_distance=tab.staffspace_height

        for index, g in enumerate(self.glyphs):

            # when a dot was found: check the position and check
            # if there is a flag in its near (left side)
            #
            # info: in french tablature a dot in course 0 can
            #       only appear right behind a flag. so, when we
            #       are in 'french mode' only dots in course 0
            #       are checked, other dots are left untouched.
            #       if a dot was found in course 0 and there was
            #       no flag in its near (i.e. it is alone), this
            #       dot will be removed.
            #       other tabtypes (e.g. italian) do not have
            #       these restrictions.
            if g.match_id_name("dot"):

                if tab.tabtype == "french" and g.course != 0:
                    continue

                # due to our sorting algorithms (dependend on
                # the pixel values of their middle), a dot
                # right behind a flag could be sorted *before*
                # that flag (in the array). this is not
                # correct and will be checked here. when the
                # dot is followed by a flag with same
                # x_number they will be swapped, so that the
                # flag is followed by 'its' dot.
                next=index+1
                if next < len(self.glyphs)\
                        and\
                        self.glyphs[next].x_number\
                        == g.x_number\
                        and\
                        self.glyphs[next].\
                        match_id_name("*flag*")\
                        and\
                        g.offset_x >\
                        self.glyphs[next].\
                        offset_x\
                        and\
                        g.offset_y >\
                        self.glyphs[next].offset_y\
                        and\
                        g.offset_y <\
                        self.glyphs[next].offset_y\
                        +self.glyphs[next].nrows:

                    tmp=self.glyphs[next]
                    self.glyphs[next]=g
                    self.glyphs[index]=tmp
                    self.glyphs[next].course=self.glyphs[index].course
                    continue

                # look for a flag that appears *before* the
                # dot
                cmp_index=index-1
                prev_x=g.x_number-1
                cur_x_number=g.x_number

                found=0
                cmp_gl=self.glyphs[cmp_index]

                while cmp_index >= 0 and\
                        cmp_gl.x_number >= prev_x and\
                        cmp_gl.x_number <= g.x_number\
                        and found == 0:

                    if cmp_gl.match_id_name("*flag*"):

                    # is the dot overlapped
                    # horizontally by a
                    # flag (i.e. x-direction)
                        if cmp_gl.offset_y < g.offset_y and\
                                cmp_gl.offset_y+cmp_gl.nrows > g.offset_y:

                            found=1

                            # dot and flag have a
                            # different x_number,
                            # although it should
                            # be equal when they
                            # are near to each
                            # other
                            if cmp_gl.x_number !=\
                                    g.x_number:

                                # compute distance
                                d=g.middle_x-cmp_gl.middle_x

                                if d < max_distance:
                                    modified=1
                                    fl_x_n=cmp_gl.x_number
                                    g.course=cmp_gl.course
                                else:
                                    found=0
                            else:
                                g.course=cmp_gl.course

                    if not found:
                        cmp_index-=1
                        cmp_gl=self.glyphs[cmp_index]

                # a flag was found in the near of the dot:
                # the modified-flag indicates, that the
                # x_numbers have to be updated for this chord.
                # all glyphs that have the same x_number as
                # the dot will get the x_number of the flag
                if modified:
                    length=len(self.glyphs)

                    # start with the first glyph behind
                    # the found flag
                    cmp_index+=1
                    #print "cmp_index:", cmp_index
                    cmp_gl=self.glyphs[cmp_index]
                    while cmp_index < length and\
                            cmp_gl.x_number\
                            <= cur_x_number:
                        cmp_gl.x_number=fl_x_n
                        cmp_index+=1
                        cmp_gl=self.glyphs[cmp_index]

                    modified=0
                    update=1

                # dots that are recognized in course 0 and
                # that are *not* part of a flag can be marked
                # here to remove them from the glyph list
                elif not found and tab.tabtype == "french":
                    rm_glyphs.append(g)

        # remove unnecessary dots
        if len(rm_glyphs) > 0:
            for rm in rm_glyphs:
                self.glyphs.remove(rm)

            update=1
            del rm_glyphs

        return update

    ######################################################################
    # Bool __is_vertical_line
    #
    # heuristic determination of a vertical line
    #
    # toom, 2004-09-07
    #
    def __is_vertical_line(self, glyph, staffline_height, staffspace_height):

        mfbhr=glyph.most_frequent_run('black','horizontal')

        if ((float(glyph.nrows)/glyph.ncols > 2.0) and \
                (glyph.ncols < mfbhr+staffline_height*0.5)) \
                or (glyph.nrows > 3.5*staffspace_height):
            return 1
        else:
            return 0
    
    def special_chords(self):
        """This method deals with some special chords using a per chord class histogram:
               repeat signs"""
        from gamera.plugins.image_utilities import union_images

        newglyphs = []
        curx = -1
        curxlist = []
        dotcount = 0
        curstaff = -1
        chordhistogram = {}

        def check_repeat():
            if chordhistogram.has_key("dot") and chordhistogram["dot"][0] > 2:
                otherglyph = []
                # deal with case that dots belonging to a repeat sign are 
                # below a flag
                for k,v in chordhistogram.iteritems():
                    if k.startswith("flag."):
                        debug_print("flag and repeat overlap", "check_repeat")
                        otherglyph.extend(v[1])


                repeatglyph = union_images(curxlist)
                repeatglyph.classify_heuristic("repeat")
                repeatglyph.gridnr = 0
                repeatglyph.x_number = curx
                repeatglyph.middle_x = (repeatglyph.ul_x + repeatglyph.lr_x) / 2
                repeatglyph.middle_y = (repeatglyph.ul_y + repeatglyph.lr_y) / 2
                repeatglyph.staff = curstaff
                repeatglyph.course = 3
                
                return [repeatglyph] + otherglyph
            else:
                return None

        # a list of methods which check the chord in curxlist and returns a list 
        # of new glyphs or None (if current chord is not a special chord).
        # some more functions could be added here
        check_functions = [check_repeat]

        for g in self.glyphs:
            if g.x_number != curx or g.staff != curstaff:
                
                changed = False
                for check_func in check_functions:
                    repeatglyphs = check_func()

                    if repeatglyphs != None:
                        newglyphs.extend(repeatglyphs)
                        changed = True
                        break

                if not changed:
                    newglyphs.extend(curxlist)
                
                chordhistogram = {}
                curxlist = []
                curstaff = g.staff
                curx = g.x_number

            curxlist.append(g)
            glyph_class = g.get_main_id()
            
            if chordhistogram.has_key(glyph_class):
                chordhistogram[glyph_class][0] += 1
                chordhistogram[glyph_class][1].append(g)
            else:
                chordhistogram[glyph_class] = [1, [g]]

                
        changed = False
        for check_func in check_functions:
            repeatglyphs = check_func()

            if repeatglyphs != None:
                newglyphs.extend(repeatglyphs)
                changed = True
                break

        if not changed:
            newglyphs.extend(curxlist)
 
        self.glyphs = newglyphs


    def barline_correction(self, tab):
        """This method tries to correct some common errors caused by
broken barlines. It is automatically called in set_glyph_properties.

It splits chords containing bars and frets into two single chords
and merges bars and other glyphs in same chord"""
        from gamera.plugins.image_utilities import union_images

        newglyphs = []
        bar = None
        merge = True
        curx = -1
        curxlist = []
        offset=0


        def build_bar_chord(curx,offset,curxlist,bar,merge):
            curx = g.x_number
            debug_print_glyphs(curxlist, "barline_correction")
            if bar != None and merge and len(curxlist) > 0:
                curxlist.append(bar)
                cc = union_images(curxlist)
                cc.classify_heuristic("bar")
                cc.gridnr = 0
                cc.x_number = bar.x_number+offset
                cc.middle_x = (cc.ul_x + cc.lr_x) / 2
                cc.middle_y = (cc.ul_y + cc.lr_y) / 2
                cc.staff = bar.staff
                cc.course = 3

                newglyphs.append(cc)
            elif bar != None: # own bar


                if len(curxlist) > 0: # bar has no own x_number

                    bar.x_number += offset
                    if bar.middle_x * len(curxlist) < sum([h.middle_x for h in curxlist]):
                        offset += 1
                        newglyphs.append(bar)
                        newglyphs.extend(curxlist)
                        for h in curxlist:
                            h.x_number += offset
                    else:
                        bar.x_number += 1
                        newglyphs.extend(curxlist)
                        newglyphs.append(bar)
                        for h in curxlist:
                            h.x_number += offset
                        offset += 1
                else:
                    bar.x_number += offset
                    newglyphs.append(bar)

            else:
                for h in curxlist:
                    h.x_number += offset

                newglyphs.extend(curxlist)

            curxlist = []
            bar = None
            merge = True
            return curx,offset,curxlist,bar,merge


        #return glyphs
        for g in self.glyphs:
            if g.x_number != curx:
                curx,offset,curxlist,bar,merge = build_bar_chord(curx,offset,curxlist,bar,merge)

            if g.match_id_name("*bar*"):
                bar = g
            elif g.match_id_name("*fret*"):
                curxlist.append(g)
                merge = False
            else:
                curxlist.append(g)
                if g.ncols > 2.5*tab.staffspace_height:
                    merge = False

        curx,offset,curxlist,bar,merge = build_bar_chord(curx,offset,curxlist,bar,merge)

        self.glyphs = newglyphs


    ######################################################################
    # classify_bars
    #
    # groups and classifies bar lines as "bar" by a heuristic method
    # useful, when many bars are broken. You can also skip this method
    # and classify bars statistically in group_list_automatic
    #
    # self.glyphs is modified so that glyphs of the same bar line
    # are grouped
    #
    # This method should be called *after* set_glyph_properties
    #
    # parameters: tab: initialized TabPage object
    #
    # chris, 2004-08-05
    #
    def classify_bars(self, tab):
        """Classifies bars heuristically.
WARNING: this function is deprecated; use TabPage.remove_barlines() instead.

Signature:

  ``classify_bars(tab)``

with

  *tab*:
    Initialized TabPage_ object where its method `remove_staves`_ has to be
    called before.

.. _TabPage: gamera.toolkits.otr.otr_staff.TabPage.html
.. _`remove_staves`: gamera.toolkits.otr.otr_staff.TabPage.html#remove-staves

*classify_bars* finds, groups (in case of broken bar lines) and classifies
bar lines heuristically. It is however preferable when bar lines are
randomly broken.  This function uses the logical position of a glyph (i.e.
*staff*, *x_number* and *course*). Therefor, this method should be called
**after** set_glyph_properties_.

.. note:: The detected bar lines will be marked as being classified
          heuristically. Do not use this method when you want to classify bar
          lines statistically with the Gamera functions
          *classify_list_automatic* or *group_list_automatic*! These
          functions will reclassify the glyphs with ``classification_state
          = HEURISTIC``.

In case some bars have been classified, the return value of this method is
``1``, otherwise ``0``.
"""

        sys.stderr.write("TabGlyhpList.classify_bars is deprecated: use TabPage.remove_bars instead!\n")
        modified=0  # indicates whether the list of glyphs has
                # been modified or not (return value)

        # do the grouping if a chord looks like a bar line
        def check_and_group_bar(chord):
            height = 0 # total height of vertical line components
            i=0
            while i < len(chord):
                # heuristic classification of vertical lines
                if self.__is_vertical_line(chord[i], \
                                tab.staffline_height, \
                        tab.staffspace_height):
                    height += chord[i].nrows
                    i+=1
                else:
                    chord.pop(i)

            if height > tab.staffspace_height*3:
                # replace parts with group in self.glyphs
                bargroup = image_utilities.union_images(chord)
                bargroup.classify_heuristic("bar")
                bargroup.staff=chord[0].staff
                bargroup.x_number=chord[0].x_number
                bargroup.middle_x=int(bargroup.offset_x+\
                        bargroup.ncols*0.5)
                bargroup.middle_y=int(bargroup.offset_y+\
                        bargroup.nrows*0.5)
                bargroup.course=chord[0].course
                for b in chord:
                    self.glyphs.remove(b)
                self.glyphs.append(bargroup)
                return 1

            return 0

        # set x_number etc. and order by it
        if not self.glyph_props_set:
            self.set_glyph_properties(tab)

        # remove all CCs outside staff ranges
        sb = tab.get_staffboxes(tightness=0)
        ccs = [c for c in self.glyphs if \
               (c.staff > 0) and \
               (c.middle_y > sb[c.staff-1].y) and \
               (c.middle_y < sb[c.staff-1].y + sb[c.staff-1].nrows)]
        #img = image_utilities.union_images(ccs)
        #img.save_tiff("union.tiff")

        # check all chords for bar candidates
        ccs.sort(self.cmp_glyph_properties)
        prev_xnum = -2
        chord = []
        for c in ccs:
            if (prev_xnum != -2) and (c.x_number != prev_xnum):
                # process previous chord
                m=check_and_group_bar(chord)
                if m:
                    modified=1
                chord = [c]
            else:
                chord.append(c)
            prev_xnum = c.x_number
        # do not forget last chord
        check_and_group_bar(chord)

        return modified

    ######################################################################
    # FloatType __get_overlap(self, a, b)
    #
    # get the percentual part of 'a' that is overlapped by 'b'. the
    # return value is between 0.0 and 1.0.
    #
    # toom, 2004-06-23
    #
    def __get_overlap(self, a, b):
        if b.offset_x <= a.offset_x:
            # no overlap
            if b.offset_x+b.ncols <= a.offset_x:
                return 0.0

            # 100% overlap
            elif b.offset_x+b.ncols >= a.offset_x+a.ncols:
                return 1.0

            # compute overlap
            else:
                length=b.offset_x+b.ncols-a.offset_x
                return (float)(length)/a.ncols
        else:
            # no overlap
            if b.offset_x >= a.offset_x+a.ncols:
                return 0.0

            # compute overlap
            elif b.offset_x+b.ncols >= a.offset_x+a.ncols:
                length=a.offset_x+a.ncols-b.offset_x
                return (float)(length)/a.ncols

            # b is within the borders of a
            else:
                return (float)(b.ncols)/a.ncols

    ######################################################################
    # FloatType __get_min_overlap(self, a, b)
    #
    # get the minimum overlap of the 2 glyphs a and b. the return value
    # is between 0.0 and 1.0.
    #
    # toom, 2004-06-23
    #
    def __get_min_overlap(self, a, b):
        o1=self.__get_overlap(a, b)
        o2=self.__get_overlap(b, a)

        #print "OVERLAP:", o1, o2
        #print "   a:", a
        #print "   b:", b

        if o1 <= o2:
            return o1
        else:
            return o2

    ######################################################################
    # FloatType __get_max_overlap(self, a, b)
    #
    # get the maximum overlap of the 2 glyphs a and b. the return value
    # is between 0.0 and 1.0.
    #
    # toom, 2004-06-23
    #
    def __get_max_overlap(self, a, b):
        o1=self.__get_overlap(a, b)
        o2=self.__get_overlap(b, a)

        #print "__get_max_overlap():", o1, o2
        #print "(       a:", a
        #print "        b:", b,  ")"

        if o1 >= o2:
            return o1
        else:
            return o2

    ######################################################################
    # sort_glyphs
    #
    # sort the list of glyphs, using the function cmp_glyph_coords (this
    # function sorts the glyph by their coordinates).
    #
    # toom, 2004-06-01
    #
    #def sort_glyphs(self):
    #   self.glyphs.sort(self.cmp_glyph_coords(0))

    #####################################################################
    #
    # string to_abctab(self)
    #
    # returns a string, formatted for use with the program abctab2ps.
    #
    # toom, 2004-07-12
    #
    def to_abctab(self, tabtype, title="OTR result", numlines=6, default_note_length=1.0/4.0):
        """Gives an output of the sorted list ``self.glyphs`` as an
*abc*-formatted string.

Signature:

  ``to_abctab(tabtype, title=OTR result, numlines=6, default_note_length=1.0/4.0)``

with

  *tabtype*:
    Determines the type of the tablature. Currently supported values are
    ``french`` and ``italian``.

  *title*:
    Value for T: header field.

  *numlines*:
    Number of Stafflines per staff

  *default_note_length*:
    Default length of notes for paramter L:

The returned string can be given as input to the external program
`abctab2ps`__.

.. __: http://www.lautengesellschaft.de/cdmm/

.. note:: The properties of the glyphs have to be set by calling
          set_glyph_properties_ before.
"""
        #
        # build the chord, using a nice output format for abc
        #
        def build_chord():
            chord_items=0
            flag_str=""

            dotflags = { '3':4, \
                     '6':8, \
                     '12':16, \
                     '24':32, \
                     '48':64, \
                     '96':128 }

            # count the items of the chord
            for i in range(len(chord)):
                if chord[i] != ',':
                    chord_items+=1

            flag = []
            if len(flags) > 0:
                flag = flags.pop(0)
            if len(flag) > 0:
                chord_items+=1
                ff = flag[1]
                if ((flag[1] % 3) == 0):
                    key = str(flag[1])
                    flag_str="3/"+str((dotflags[key]/div))

                elif flag[0] == 1:
                    if flag[1] > div:
                        while (ff > div):
                            flag_str+='/'
                            ff/=2
                    else:
                        flag_str=str(int(div/flag[1]))
                else:
                    ff/=div
                    flag_str=str(flag[0])+"/"+str(ff)

            if dot != "":

                chord_items+=len(dot)

            if len(flags) == 0:
                return dot + "[" + chord + flag_str + "] ", chord_items

            return dot + "[" + chord + flag_str + "]", chord_items

        hasbeams=False
        chord=""
        course=1
        cur_staff=1
        cur_x_number=1
        flags=[]
        dot=""
        n_items=0
        n_chords=0
        song=""
        div=1
        last_bar=0
        dotcount = 0
        flagcourse = -1

        if tabtype == "french":
            if numlines == 5:
                tabtype="french5tab"
            elif numlines == 4:
                tabtype="french4tab"
            else:
                tabtype="frenchtab"
        elif tabtype == "italian":
            tabtype="italiantab"

        decomap = {"hash": "#", "star": "*"}
        mensuralmap = {"c": "[M:C]", "c_cut": "[M:C|]", "3": "[M:3]"}

        flag_course=-1
        for i, g in enumerate(self.glyphs):

            # ignore characters outside of the staves
            if g.staff < 0 or g.course < 0:
                continue

            # start a new staff
            if g.staff != cur_staff:
                if chord != "":
                    # for output reasons: print every bar
                    # in an own line
                    if last_bar != 0:
                        song+="\\\n\t"

                    chord, no=build_chord()
                    n_items+=no
                    n_chords+=1
                    song+=chord
                    chord=""

                last_bar=0
                n_staff=g.staff-cur_staff

                song+="\n%\n"
                n_staff-=1
                flags = []

                # comment this here, if empty staves are not
                # wanted (e.g. if no symbols are in the
                # 3rd staff, then it would be printed though)
                #while n_staff > 0:
                #   song+="y\n%\n"
                #   n_staff-=1

                cur_staff=g.staff

            if g.x_number != cur_x_number:
                debug_print("g.x != cur.x", "to_abctab")
                if chord != "":
                    # for readability: print each bar in its own line
                    if last_bar != 0:
                        song+="\\\n\t"
                        last_bar=0

                    chord, no=build_chord()
                    n_items+=no
                    n_chords+=1
                    song+=chord
                    chord=""

                course=1
                dot=""
                cur_x_number=g.x_number

            #
            # test for recognized glyphs
            #
            if g.match_id_name("bar"):
                song+="|"
                n_items+=1
                n_chords+=1

                # remember the last bar (for producing nice
                # formatted output)
                last_bar=g.staff

                # produce nice formatted output
                #if i < len(self.glyphs)-1:
                #   if self.glyphs[i+1].staff == cur_staff:
                #       song+="\\\n\t"
            elif g.match_id_name("repeat"):
                if song[-3:] == ":||":
                    song = song[:-3] + "::"
                elif song[-2:] == ":|":
                    song = song[:-2] + "::"
                elif song[-2:] == "::":
                    pass
                else:
                    song += ":"
            elif g.match_id_name("*flag.*.*"):
                f = g.id_name[0][1].split('.')[1:]

                #remove incomplete flags from last runs
                if len(flags) > 0:
                    song += " " #"[y]"*len(flags) + " "
                    flags = []
                i = 0
                while i < len(f):
                    debug_print((i, f[i]), "to_abctab")
                    if f[i].startswith("dot"):
                        dotidx = int(f[i][3:])
                        if dotidx < len(flags):
                            #denom = flags[-dotidx][0]
                            #flags[-dotidx][0] = flags[-dotidx][0]*2+denom
                            flags[dotidx-1][0] *= 3
                            flags[dotidx-1][1] *= 2
                        debug_print("has dot and index %d" % dotidx, "to_abctab")
                        i += 1
                    else:
                        debug_print((i, f[i], f[i+1]), "flags")
                        for j in xrange(int(f[i+1])):
                            flags.append([1,int(f[i])])

                        i += 2

                debug_print((flags), "flags")

                hasbeams = True

            elif g.match_id_name("*flag*"):
                if len(flags) > 0:
                    song += " " # "[y]"*len(flags) + " "
                    flags = []

                #if len(flags) > 0:
                #    print "??????????????????????????????????????", len(flags), flags, g.staff, g.x_number
                # get the value of the flag and calculate
                # its number for the abctab-output
                if flagcourse < 0:
                    flagcourse = g.course
                f=g.id_name[0][1].split('.')
                if f[-1] != "longa":
                    speed=int(f[-1])
                else:
                    speed = 1
                flags=[[1,speed]]
                flag_course=g.course

            elif g.match_id_name("*fret*"):
                while course < g.course:
                    chord+=","
                    course+=1
                f=g.id_name[0][1].split('.')
                chord+=f[-1]
                course+=1

            elif g.match_id_name("*dot*"):
                debug_print("DOT", "to_abctab")
                if len(flags)>0 and\
                        g.gridnr <= len(flags) and g.course == flagcourse:
                        #                    g.gridnr > 0 and
                    #                        and g.course == flag_course:
                    if g.gridnr > 0: #dotgridnr
                        debug_print(( "B", flags), "to_abctab")
                        flags[g.gridnr-1][0] *= 3
                        flags[g.gridnr-1][1] *= 2
                    else:
                        debug_print(( "A", flags), "to_abctab")
                        flags[g.gridnr][0] *= 3
                        flags[g.gridnr][1] *= 2
                else:
                    debug_print(("C", i, g.x_number, g.gridnr, g.course, flag_course, flags), "to_abctab")
                    dot+="."

            elif g.match_id_name("fermata") or g.match_id_name("deco.fermata"):
                dot += "H"

            elif g.match_id_name("deco.*.here"):
                while course < g.course:
                    chord+=","
                    course+=1
                try:
                    chord += decomap[g.get_main_id().split(".")[1]] + "y"
                except:
                    debug_print("unknown deco %s" % g.get_main_id(), "error")
                    print "skipping symbol"
                course+=1
            elif g.match_id_name("deco.*.left") or g.match_id_name("deco.*.right"):
                while course < g.course:
                    chord+=","
                    course+=1

                #get corresponding fret
                cmp_gl = self.glyphs[i+1]
                debug_print((cmp_gl.x_number, g.x_number, cmp_gl.course, g.course, cmp_gl.staff, g.staff), "deco")
                if cmp_gl.x_number == g.x_number and cmp_gl.course == g.course and cmp_gl.staff == g.staff:
                    chord = chord[:-1] + decomap[g.get_main_id().split(".")[1]] + chord[-1]
                    debug_print((chord), "deco")

                elif i > 1:
                    cmp_gl = self.glyphs[i-1]
                    debug_print((cmp_gl.x_number, g.x_number, cmp_gl.course, g.course, cmp_gl.staff, g.staff), "deco")
                    if cmp_gl.x_number == g.x_number and cmp_gl.course == g.course and cmp_gl.staff == g.staff:
                        chord = chord[:-1] + decomap[g.get_main_id().split(".")[1]] + chord[-1]
                        debug_print((chord), "deco")

            elif g.match_id_name("time.*"):
                try:
                    timecode = g.get_main_id().split(".")[1]
                    song += " %s " % mensuralmap[timecode]
                except KeyError:
                    debug_print(("unknown time signature", g, g.id_name, g.x_number, g.staff), "error")
            elif g.match_id_name("trash"):
                pass
            else:
                debug_print(("unknown symbol", g, g.id_name, g.x_number, g.staff), "error")




        # don't forget the last chord
        debug_print(("remaining chord", chord), "to_abctab")
        if chord != "":
            # for readability: print each bar in its own line
            if last_bar != 0:
                song+="\\\n\t"
                last_bar=0

            chord, no=build_chord()
            n_items+=no
            n_chords+=1
            song+=chord

        debug_print(("remaining flags", flags), "to_abctab")
        # statistics as comment
        song+="\n% number of items: " + str(n_items) + "\n"


        # add header fields
        #TODO: case if no beams present
        song = "X:1\n"+\
               "%\n"+\
               "%%titleleft true\n"+\
               "%%tabaddflags 0\n"+\
               "%%tabrhstyle grid\n"+\
               "%\n"+\
               "T:" + title + "\n"+\
               "C:" + str(n_items) + " symbols in " + str(n_chords) + " chords\n"+\
               "L:1/"+str((int)(1.0/default_note_length))+"\n"+\
               "K:" + str(tabtype) + "\n"+\
               "%\n" +\
               song


        return song

    #####################################################################
    #
    # string to_abcmusic(self,tuning,key,factor)
    #
    # returns a string in abc music format
    # tuning is used for transcription tab -> music
    # key is for the key signature, eg. Eb for "E flat Major"
    # or Cm for "C Minor", key='auto' tries to guess the key
    # factor is a logarithmic length factor for slowing down,
    # eg. factor=2 outputs quarter notes for sixteenth notes etc
    #
    # chris, 2004-09-15
    #
    def to_abcmusic(self, tuning="gdAFCG,", key='auto', factor=0):
        """Gives an output of the sorted list ``self.glyphs`` as an
*abc music*-formatted string.

Signature:

  ``to_abcmusic(tuning=\"gdAFCG,\", key=\"auto\", factor=0)``

with

  *tuning*:
    The tuning of each course, from top to bottom. Each note is specified
    as *abc* code (sharp = prefix '^', flat = prefix '_', octave up = postfix
    single quote, octave down = postfix comma)

  *key*:
    Key signature, e.g. the value ``Eb`` stands for *E flat Major* or ``Cm``
    for *C Minor*. When set to ``auto`` the key is guessed automatically.

  *factor*:
    Logarithmic length factor for slowing down. For example, when set to
    ``2`` the sixteenth notes are printed as quarter notes. A negative
    integer leads to a speedup.

The returned string can be given as input to the external program
`abctab2ps`__.

.. __: http://www.lautengesellschaft.de/cdmm/

.. note:: The properties of the glyphs have to be set by calling
          set_glyph_properties_ before.
"""
        # some settings for tab -> music transcription
        abcmusic=ToAbcmusic(self.glyphs,"french",tuning,key,factor)
        song=abcmusic.abc_music

        return song

    #####################################################################
    # string to_diff_output
    #
    # for testing only
    #
    # toom, 2004-09-16
    def to_diff_output(self):
        glyph_information=""
        for g in self.glyphs:
            if g.course >= 0 and g.staff >= 0:
                glyph_information+=str(g.id_name[0][1])
                if not g.match_id_name('*bar*'):
                    glyph_information+=" "+str(g.course)
                glyph_information+='\n'

        return glyph_information

    #####################################################################
    # IntType cmp_glyph_properties(self, a, b)
    #
    # compare function for sorting the glyphs as follows (ascending):
    #
    #       1. staff
    #       2. x_number
    #       3. course
    #
    # toom, 2004-06-23
    #
    def cmp_glyph_properties(self, a, b):
        staffDiff=a.staff-b.staff
        if staffDiff != 0:
            return staffDiff

        x_number_diff=a.x_number-b.x_number
        if x_number_diff != 0:
            return x_number_diff
#        if a.gridnr >= 0 and b.gridnr >= 0


        course_diff = a.course-b.course
        if course_diff != 0:
            return course_diff

        gridnrdiff = a.gridnr-b.gridnr
        if gridnrdiff != 0:
            return gridnrdiff

        # when everything identical, sort by x coordinate
        return a.ul_x-b.ul_x

    def cmp_glyph_properties_grid(self, a, b):
        staffDiff=a.staff-b.staff
        if staffDiff != 0:
            return staffDiff

        x_number_diff=a.x_number-b.x_number
        if x_number_diff != 0:
            return x_number_diff
#        if a.gridnr >= 0 and b.gridnr >= 0

        gridnrdiff = a.gridnr-b.gridnr
        if gridnrdiff != 0:
            return gridnrdiff

        course_diff = a.course-b.course
        if course_diff != 0:
            return course_diff


        # when everything identical, sort by x coordinate
        return a.ul_x-b.ul_x



    #####################################################################
    # IntType cmp_glyph_coords(self, a, b)
    #
    # parameters: a, b: TabGlyph objects
    #
    # compare function for sorting the glyphs as follows:
    #
    #           1. staff number
    #           2. x-position in the same staff
    #           3. y-position in the same staff
    #
    # needs to be defined as a callable object so that a tolerance
    # can be passed as an additional argument
    #
    # 2004-08-02, chris
    #
    class cmp_glyph_coords:
        def __init__(self, tolerance=0):
            self.__tolerance = tolerance

        def __call__(self, a, b):
            # if the glyphs are on different staffs
            staffDiff=a.staff-b.staff
            if staffDiff != 0:
                return staffDiff

            if a.middle_x == b.middle_x:
                return a.middle_y-b.middle_y

            return a.middle_x-b.middle_x

        def __call_old__(self, a, b):
            # if the glyphs are on different staffs
            staffDiff=a.staff-b.staff
            if staffDiff != 0:
                return staffDiff

            if self.__tolerance == 0:
                halfwidth_a=a.ncols/2
                halfwidth_b=b.ncols/2
                if halfwidth_b > halfwidth_a:
                    self.__tolerance=halfwidth_b
                else:
                    self.__tolerance=halfwidth_a

            # same x-position? should be +/- a value, because the
            # x-positions are *never* the same for different
            # characters
            if abs(a.middle_x-b.middle_x) > self.__tolerance:
                return a.middle_x-b.middle_x
            else:
                return a.middle_y - b.middle_y

    def split_diapasons(self, ratio_threshold=2.4):
        """This method splits glyphs with an aspect ratio (ncols/nrows) > *ratio_threshold*
into ncols/nrows parts.

It can be used to split connected diapasons and should be called before the glyphs are classified
"""

        def splitdiapason(cc):
            ratio = cc.ncols/cc.nrows
            centers = [1.0/ratio*i for i in range(1,ratio)]
            return cc.splitx(centers)

        added = []
        removed = []
        for g in self.glyphs:
            if float(g.ncols)/g.nrows > ratio_threshold:
                debug_print("splitting diapason")
                added.extend(splitdiapason(g))
                removed.append(g)

        for g in removed:
            self.glyphs.remove(g)

        self.glyphs.extend(added)


