#!/usr/bin/python

#
# otr_recognize - recognizes a French/Italian tablature image
#                 and creates abc output
#                 this file is part of the OTR toolkit for gamera
#
# Authors:  Christoph Dalitz  2004-2013
#           Thomas Karsten    2004
#           Christian Brandt  2012
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 

# initialization
import sys
import os
import re
import os.path

from gamera.core import *
from gamera.config import config
from gamera import knn
from gamera.classify import BasicGroupingFunction, ShapedGroupingFunction

from gamera.toolkits.otr.otr_staff import *
from gamera.toolkits.otr.otr_glyph import *
from gamera.toolkits.otr.plugins import *

#
# parse command line arguments
#
class c_opt:
	infile = ""
	outfile = ""
	database = ""
	verbose = 1
	onlines = 0
	nlines = 0
	num_k = 1
	tabtype = "french"
	smooth = 0
	deskew = 0
	depict = 0
	despec = 0
	knnbars = 0
	abcmusic = 0
        splitdiapason = 0
	tuning = "gdAFCG,"
	def error_exit(self):
		usage = "Usage:\n\t" + sys.argv[0] + " [<options>] <infile>\n" + \
				"Options:\n" + \
				"\t-o <abc>  abc output file name ([basename + '.abc'])\n" + \
				"\t-d <xml>  training data file ['" + self.database + "']\n" + \
				"\t-n <num>  number of staff lines (autoguessed when omitted)\n" + \
				"\t-k <k>    value of k in kNN ([" + ("%d" % opt.num_k) +"])\n" + \
				"\t-v <lev>  verbose level ([" + ("%d" % opt.verbose) + "])\n" + \
				"\t-smooth   perform smoothing\n" + \
				"\t-deskew   perform deskewing\n" + \
				"\t-depict   remove pictures\n" + \
				"\t-despec   remove CC's smaller than staffline_height * 2\n" + \
				"\t-onlines  fret symbols on the lines\n" + \
				"\t-tabtype  indicates the tablature type\n" + \
				"\t          possible values: 'french', 'italian' ['" + self.tabtype  + "']\n" + \
				"\t-split    splits diapasons\n" + \
				"\t-knnbars  do not recognize bars heuristically but\n" + \
				"\t          statistically with knn-Classifier\n" + \
				"\t-abcmusic write music transcription in addition to tabcode\n" + \
				"\t-tuning <tuning>\n" + \
				"\t          tuning for music transcription, given from top\n" + \
				"\t          to bottom. Default is \"" + self.tuning + "\"\n"
		sys.stderr.write(usage)
		sys.exit(1)
opt = c_opt()

i = 1
while i < len(sys.argv):
	if sys.argv[i] == "-o":
		i += 1; opt.outfile = sys.argv[i]
	elif sys.argv[i] == "-d":
		i += 1; opt.database = sys.argv[i]
	elif sys.argv[i] == "-n":
		i += 1; opt.nlines = int(sys.argv[i])
	elif sys.argv[i] == "-k":
		i += 1; opt.num_k = int(sys.argv[i])
	elif sys.argv[i] == "-v":
		i += 1; opt.verbose = int(sys.argv[i])
	elif sys.argv[i] == '-smooth':
		opt.smooth = 1
	elif sys.argv[i] == '-deskew':
		opt.deskew = 1
	elif sys.argv[i] == '-depict':
		opt.depict = 1
	elif sys.argv[i] == '-despec':
		opt.despec = 1
	elif sys.argv[i] == '-onlines':
		opt.onlines = 1
	elif sys.argv[i] == '-split':
		opt.splitdiapason = 1
	elif sys.argv[i] == '-tabtype':
		i += 1; opt.tabtype = sys.argv[i]
	elif sys.argv[i] == '-knnbars':
		opt.knnbars = 1
	elif sys.argv[i] == '-abcmusic':
		opt.abcmusic = 1
	elif sys.argv[i] == '-tuning':
		i += 1; opt.tuning = sys.argv[i]
		if re.search("[^_^,'a-gA-G]", opt.tuning):
			sys.stderr.write("Syntax error in tuning \"" + opt.tuning + "\"\n")
			opt.error_exit()
	elif sys.argv[i][0] == '-':
		opt.error_exit()
	else:
		opt.infile = sys.argv[i]
	i += 1

if (opt.infile == "") or not os.path.isfile(opt.infile):
	sys.stderr.write("File '" + opt.infile + "' not found\n")
	opt.error_exit()

if (opt.outfile == ""):
	opt.outfile = re.sub("\.[^.]*$","",os.path.basename(opt.infile)) + ".abc"

# prevent gamera from parsing arguments
sys.argv = []

init_gamera()

if opt.verbose >= 1:
	config.set("progress_bar",True)

#
# load an image, create a TabStaff object and remove staves
#
image=load_image(opt.infile)

# do quested preprocessing
if opt.deskew:
	if image.data.pixel_type != GREYSCALE:
		image = image.to_greyscale()
	image = image.deskew()

if opt.smooth:
	if image.data.pixel_type != GREYSCALE:
		image = image.to_greyscale()
	image = image.smoothen()

if image.data.pixel_type != ONEBIT:
	image = image.to_onebit()

if opt.depict:
	if opt.nlines != 0:
            image = image.remove_pictures(opt.nlines)
        else:
            image = image.remove_pictures(5)

if opt.despec:
	size = image.most_frequent_run('black','vertical') * 2
	image.despeckle(size)
	
tab=TabPage(image, opt.onlines, opt.tabtype)
if opt.onlines:
	tab.remove_staves(crossing_symbols='all', num_lines=opt.nlines)
else:
	tab.remove_staves(crossing_symbols='bars', num_lines=opt.nlines)
#tab.image.save_tiff("%s-nostaff.tif" % re.sub("\.[^.]*$","",os.path.basename(opt.infile)))
#sys.exit(0)

barlines = []
if not opt.knnbars:
	barlines = tab.remove_barlines()
# 	rgb = tab.image.to_rgb()
# 	for c in barlines:
# 		rgb.highlight(c, RGBPixel(255, 0, 0))
# 	rgb.save_PNG("bars.png")

# remove small remnants from staff removal
tab.image.despeckle(3)

# get all connected components and create a TabGlyphList object
ccs=tab.image.cc_analysis()
glyphs=TabGlyphList(ccs + barlines)


#
# create a classifier and load the database
#
classifier=knn.kNNInteractive([], [\
		'aspect_ratio',\
		'moments',\
		'nrows_feature',\
		'volume64regions',\
		], 0)
classifier.num_k = opt.num_k
classifier.from_xml_filename(opt.database)

# split touching diapasons, if requested
if opt.splitdiapason:
	glyphs.split_diapasons()

#
# classify the connected components of the image, use fraction of staffspace
# height to decide about grouped glyphs
#
grp_distance=tab.staffspace_height/3
added, removed=classifier.group_list_automatic([c for c in glyphs.glyphs\
		if c.classification_state == UNCLASSIFIED],\
		ShapedGroupingFunction(grp_distance),max_parts_per_group=3,criterion='avg')

#
# detected parts of glyphs and special (useless) glyphs will be removed
#
if not opt.knnbars:
	glyphs.glyphs=[x for x in glyphs.glyphs if\
		(not x.match_id_name("_group._part.*")) and\
		(not x.match_id_name("*trash*")) and\
		(x.nrows > tab.staffline_height) and\
		(not (x.match_id_name("*bar*") and x.classification_state != 2)) ]
else:
	glyphs.glyphs=[x for x in glyphs.glyphs if\
		(not x.match_id_name("_group._part.*")) and\
		(not x.match_id_name("*trash*")) and\
		(x.nrows > tab.staffline_height)]

#
# add both the glyphs that were detected as a group and the manually
# classified glyphs
#
if len(added) > 0:
	if opt.verbose >= 2:
		print "Added:", added
	glyphs.glyphs.extend(added)

#
# set the properties of each glyph and convert to .abc
#
glyphs.set_glyph_properties(tab)
if opt.verbose >= 2:
	print "---------------------------------------------"
	for g in glyphs.glyphs:
		print g.id_name, "staff:", g.staff, "xpos:", g.x_number,
		print "course:", g.course
		print "   ncols:", g.ncols, "nrows:", g.nrows

#print tab.get_staffboxes()
song=glyphs.to_abctab(tab.tabtype, numlines=opt.nlines)
if opt.verbose >= 2:
	print "---------------------------------------------"
	for g in glyphs.glyphs:
		print g, g.middle_x, g.middle_y, g.id_name[0][1], "staff:", g.staff, "xpos:", g.x_number, "course:", g.course
		#g.nrows, g.middle_x

# set title to filename
song = re.sub("\nT:.*\n",("\nT:%s\n" % os.path.basename(opt.infile)),song,1)

abc_file=open(opt.outfile, "w")
abc_file.write(song)
abc_file.write("\n\n")
if opt.abcmusic:
	abc_file.write(glyphs.to_abcmusic(tuning=opt.tuning,factor=1))
abc_file.close()

#os.system("abcselect -X 2 " + opt.outfile + " > music.abc && " + \
#	  "abc2midi music.abc -o music.mid && open music.mid")
